import traceback
import re
from usdm_excel.cdisc_ct import CDISCCT
from usdm_excel.globals import Globals

class QuantityType():

  def __init__(self, quantity_info: str, globals: Globals, allow_missing_units: bool=True, allow_empty: bool=True):
    try:
      self.value = None
      self.units = None
      self.units_code = None
      self.errors = []
      self.empty = False
      self.label = quantity_info.strip()
 
      if quantity_info:
        match = re.match(r"(?P<value>[+|-]*\d+)(\s*(?P<units>.+))?", self.label)
        if match :
          parts = match.groupdict()
          self.value = parts['value'].strip()
          if parts['units']:
            self.units = parts['units'].strip()
            self.units_code = CDISCCT(globals).code_for_unit(self.units)
            if not self.units_code:        
              self.errors.append(f"Unable to set the units code for the quantity '{quantity_info}'")
          elif allow_missing_units:
            pass
          else:
            self.errors.append(f"Could not decode the quantity value, possible typographical errors '{quantity_info}'")
        else:
          self.errors.append(f"Could not decode the quantity value '{quantity_info}'")
      elif not allow_empty:
        self.errors.append(f"Could not decode the quantity value, appears to be empty '{quantity_info}'")
      else:
        self.empty = True
    except Exception as e:
      self.errors.append(f"Exception '{e}' raised decoding quantity '{quantity_info}")




from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .code import Code
from .comment_annotation import CommentAnnotation
from .transition_rule import TransitionRule

class Encounter(ApiBaseModelWithIdNameLabelAndDesc):
  type: Code
  previousId: Union[str, None] = None
  nextId: Union[str, None] = None
  scheduledAtId: Union[str, None] = None
  environmentalSettings: List[Code] = []
  contactModes: List[Code] = []
  transitionStartRule: Union[TransitionRule, None] = None
  transitionEndRule: Union[TransitionRule, None] = None
  notes: List[CommentAnnotation] = []
  instanceType: Literal['Encounter']



import json
import stringcase
from uuid import uuid4

class Neo4jDict():

  class LogicError(Exception):
    pass
  
  def __init__(self, study):
    self.study = study
    self.nodes = {}
    self.edges = {}
    self.add_edges = []
    self.node_id_to_uuid_map = {}
      
  def to_dict(self):
    node = json.loads(self.study.to_json_with_type())
    self._process_node(node)
    for edge in self.add_edges:
      if edge['end'] in self.node_id_to_uuid_map:
        self._add_edge(edge['start'], self.node_id_to_uuid_map[edge['end']], edge['raw_relation'])
      else:
        raise self.LogicError(f"{edge['start']} --edge-> {edge['end']} [{edge}]")
    return {'nodes': self.nodes, 'edges': self.edges}
  
  def _process_node(self, node):
    if type(node) == list:
      if node:
        return [self._process_node(item) for item in node]
      else:
        return None
    elif type(node) == dict and '_type' not in node:
      return str(node)
    elif type(node) == dict:
      properties = {}
      this_node_uuid = str(uuid4())
      for key, value in node.items():
        if self._is_edge_field(key):
          self._edge_field(key, value, this_node_uuid)
        else:
          result = self._process_node(value)
          if result == None:
            pass
          elif isinstance(result, list):
            if isinstance(result[0], str):
              properties[key] = value
            else:
              for item in result:
                self._add_edge(this_node_uuid, item['uuid'], key)
          elif isinstance(result, dict):
            self._add_edge(this_node_uuid, result['uuid'], key)
          else:
            properties[key] = value
      self._add_node(node['_type'], this_node_uuid, properties)
      self.node_id_to_uuid_map[properties['id']] = this_node_uuid
      return properties
    else:
      return node

  def _add_node(self, klass, uuid, properties):
    if klass == "Study":
      properties['id'] = uuid
      #print(f"Properties: {properties}")
    if klass not in self.nodes:
      self.nodes[klass] = []
    properties.pop('_type')
    properties['uuid'] = uuid
    self.nodes[klass].append(properties)

  def _is_edge_field(self, key):
    if key == "conditionAssignments" or key.endswith('Ids') or key.endswith('Id'):
      return True
    return False

  def _edge_field(self, key, value, current_index):
    # if key == "conditionAssignments":
    #   for item in value:
    #     self._add_post_edge(current_index, item['conditionTargetId'], key)
    #elif key.endswith('Ids'):
    if key.endswith('Ids'):
      for item in value:
        if item:
          self._add_post_edge(current_index, item, key)
    elif key.endswith('Id'):
      if value:
        self._add_post_edge(current_index, value, key)

  def _add_post_edge(self, start, end, key):
    self.add_edges.append({'start': start, 'end': end, 'raw_relation': key})

  def _add_edge(self, start, end, key):
    rel = self._rel_name(key)
    name = f"{stringcase.snakecase(rel).upper()}_REL"
    if name not in self.edges:
      self.edges[name] = []
    self.edges[name].append({'start': start, 'end': end})

  def _rel_name(self, key):
    if key == "conditionAssignments":
      return key
    elif key.endswith('Ids'):
      return key[:-3]
    elif key.endswith('Id'):
      return key[:-2]
    return key


from typing import Literal
from .api_base_model import ApiBaseModelWithId
from .code import Code

class ResponseCode(ApiBaseModelWithId):
  isEnabled: bool
  code: Code
  instanceType: Literal['ResponseCode']



from usdm_excel.option_manager import *

def test_create(globals):
  object = OptionManager(globals)
  assert len(object._items.keys()) == 0
  assert object._items == {}

def test_set(globals):
  globals.option_manager._items = {}
  globals.option_manager.set('fred', 'value')
  assert len(globals.option_manager._items.keys()) == 1
  assert globals.option_manager._items['fred'] == 'value'

def test_get(globals):
  globals.option_manager._items = {}
  globals.option_manager._items['fred'] = 'value'
  assert globals.option_manager.get('fred') == 'value'

def test_clear(globals):
  globals.option_manager._items = {}
  globals.option_manager._items['fred'] = 'value'
  assert len(globals.option_manager._items.keys()) == 1
  globals.option_manager.clear()
  assert len(globals.option_manager._items.keys()) == 0

def test_options(globals):
  globals.option_manager.set(Options.EMPTY_NONE, EmptyNoneOption.NONE)
  assert globals.option_manager.get(Options.EMPTY_NONE) == EmptyNoneOption.NONE.value
  globals.option_manager.set(Options.EMPTY_NONE, EmptyNoneOption.EMPTY)
  assert globals.option_manager.get(Options.EMPTY_NONE) == EmptyNoneOption.EMPTY.value



import pytest
import pandas as pd
from usdm_excel.study_design_activity_sheet.study_design_activity_sheet import StudyDesignActivitySheet
from usdm_model.code import Code
from tests.test_factory import Factory

xfail = pytest.mark.xfail

def test_create(mocker, globals):
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['ActivityId_1', 'ActivityId_2', 'ActivityId_3']
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [['Activity 1', 'Activity One', '', ''], ['Activity 2', 'Activity Two', 'T', 'Condition 2'], ['Activity 3', 'Activity Three', 'F', '']]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['activityName', 'activityDescription', 'activityIsConditional', 'activityIsConditionalReason'])
  activities = StudyDesignActivitySheet("", globals)
  assert len(activities.items) == 3
  assert activities.items[0].id == 'ActivityId_1'
  assert activities.items[0].name == 'Activity 1'
  assert activities.items[0].description == 'Activity One'
  assert activities.items[1].id == 'ActivityId_2'
  assert activities.items[1].description == 'Activity Two'
  assert activities.items[2].id == 'ActivityId_3'
  
def test_create_empty(mocker, globals):
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['studyActivityName', 'studyActivityDescription', 'studyActivityType'])
  activities = StudyDesignActivitySheet("", globals)
  assert len(activities.items) == 0

  



from typing import Literal, Union, List
from .api_base_model import ApiBaseModelWithId
from .code import Code
from .comment_annotation import CommentAnnotation

class StudyAmendmentImpact(ApiBaseModelWithId):
  type: Code
  text: str
  isSubstantial: bool
  notes: List[CommentAnnotation] = []
  instanceType: Literal['StudyAmendmentImpact']



from typing import Literal
from .api_base_model import ApiBaseModelWithId
from .code import Code

class Identifier(ApiBaseModelWithId):
  text: str
  scopeId: str
  instanceType: Literal['Identifier']

class ReferenceIdentifier(Identifier):
  type: Code
  instanceType: Literal['ReferenceIdentifier']

class StudyIdentifier(Identifier):
  instanceType: Literal['StudyIdentifier']



from typing import Literal, Union, List
from .api_base_model import ApiBaseModelWithId
from .code import Code

class Address(ApiBaseModelWithId):
  text: Union[str, None] = None
  lines: List[str] = []
  city: Union[str, None] = None
  district: Union[str, None] = None
  state: Union[str, None] = None
  postalCode: Union[str, None] = None
  country: Union[Code, None] = None
  instanceType: Literal['Address']



import logging
from datetime import date
from uuid import UUID
from usdm_db.errors_and_logging.errors_and_logging import ErrorsAndLogging

class CrossReference():

  def __init__(self, study, errors_and_logging: ErrorsAndLogging):
    self._errors_and_logging = errors_and_logging
    self._study = study
    self._references = {}
    self._logger = logging.getLogger(__name__)
    self._process_node(self._study)

  def get(self, klass, id):
    key = self._key(klass, id)
    if key in self._references:
      return self._references[key]
    else:
      #print(f"REFERENCES: {list(self._references.keys())}")
      return None

  def _process_node(self, node):
    if type(node) == list:
      if node:
        for item in node:
          self._process_node(item) 
    elif type(node) == str:
      pass
    elif type(node) == float:
      pass
    elif type(node) == date:
      pass
    elif type(node) == bool:
      pass
    elif type(node) == UUID:
      pass
    elif node is None:
      pass
    else:
      if hasattr(node, 'instanceType'):
        key = self._key(node.instanceType, node.id)
        self._references[key] = node
      for name, field in node.model_fields.items():
        self._process_node(getattr(node, name))
  
  def _key(self, klass, id):
    klass_name = self._klass_name(klass)
    return f"{klass_name}.{id}"

  def _klass_name(self, klass):
    return klass if isinstance(klass, str) else klass.__name__



import traceback
from usdm_excel.base_sheet import BaseSheet
from usdm_model.syntax_template_dictionary import SyntaxTemplateDictionary, ParameterMap
from usdm_excel.globals import Globals

class StudyDesignDictionarySheet(BaseSheet):

  SHEET_NAME = 'dictionaries'

  def __init__(self, file_path: str, globals: Globals):
    try:
      self.items = []
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME, optional=True)
      if self.success:
        current_name = None
        current_dictionary = None
        current_map = []
        for index, row in self.sheet.iterrows():
          name = self.read_cell_by_name(index, 'name')
          if name:
            if name != current_name:
              current_name = name
              if current_dictionary:
                current_dictionary.parameterMaps = current_map
                current_map = []
              description = self.read_cell_by_name(index, 'description')
              label = self.read_cell_by_name(index, 'label')
              current_dictionary = self._dictionary(name, description, label)
          key = self.read_cell_by_name(index, 'key')
          klass = self.read_cell_by_name(index, 'class', default="")
          xref_name = self.read_cell_by_name(index, 'xref', default="")
          attribute_path = self.read_cell_by_name(index, ['attribute', 'path'], default="")
          value = self.read_cell_by_name(index, 'value', default="", must_be_present=False)
          if klass:
            try:
              instance, attribute = self.globals.cross_references.get_by_path(klass, xref_name, attribute_path)
            except Exception as e:
              instance = None
              col = self.column_present(['attribute', 'path'])
              self._error(index, col, str(e))
            if instance:
              map = self.create_object(ParameterMap, {'tag': key, 'reference': f'<usdm:ref klass="{instance.__class__.__name__}" id="{instance.id}" attribute="{attribute}"></usdm:ref>'})
              if map:
                current_map.append(map)
          else:
            map = self.create_object(ParameterMap, {'tag': key, 'reference': f'{value}'})
            if map:
              current_map.append(map)
        # Clean up last dictionary if present
        if current_dictionary:
          current_dictionary.parameterMaps = current_map
        
    except Exception as e:
      self._sheet_exception(e)

  def _dictionary(self, name, description, label):
    try:
      item = SyntaxTemplateDictionary(
        id=self.globals.id_manager.build_id(SyntaxTemplateDictionary), 
        name=name,
        description=description,
        label=label,
        parameterMaps=[]
      )
    except Exception as e:
      self._general_exception(f"Failed to create SyntaxTemplateDictionary object", e)
      return None
    else:
      self.items.append(item)
      self.globals.cross_references.add(name, item)
      return item



from typing import Union, List, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .code import Code
from .masking import Masking
from .assigned_person import AssignedPerson

class StudyRole(ApiBaseModelWithIdNameLabelAndDesc):
  code: Code
  appliesToIds: List[str] = []  # Going to allow for an mpty list but really a role 
                                # should be attached to a Study Version or Design
  assignedPersons: List[AssignedPerson] = []
  organizationIds: List[str] = []
  masking: Union[Masking, None] = None
  instanceType: Literal['StudyRole']



from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc, ApiBaseModelWithId

class ConditionAssignment(ApiBaseModelWithId):
  condition: str
  conditionTargetId: str
  instanceType: Literal['ConditionAssignment']

class ScheduledInstance(ApiBaseModelWithIdNameLabelAndDesc):
  defaultConditionId: Union[str, None] = None
  epochId: Union[str, None] = None
  instanceType: Literal['ScheduledInstance']

class ScheduledActivityInstance(ScheduledInstance):
  timelineId: Union[str, None] = None
  timelineExitId: Union[str, None] = None
  activityIds: List[str] = []
  encounterId: Union[str, None] = None
  instanceType: Literal['ScheduledActivityInstance']

class ScheduledDecisionInstance(ScheduledInstance):
  conditionAssignments: List[ConditionAssignment] = [] # Allow for empty list, not in API
  instanceType: Literal['ScheduledDecisionInstance']



from usdm_model.wrapper import Wrapper
from usdm_model.study import Study
from usdm_model.study_design import StudyDesign
from usdm_model.study_version import StudyVersion
from usdm_model.study_title import StudyTitle
from usdm_model.study_definition_document import StudyDefinitionDocument
from usdm_model.study_definition_document_version import StudyDefinitionDocumentVersion
from usdm_model.code import Code
from usdm_model.identifier import StudyIdentifier
from usdm_model.organization import Organization
from usdm_model.address import Address
from usdm_model.narrative_content import NarrativeContent
from usdm_db.errors_and_logging.errors_and_logging import ErrorsAndLogging
from usdm_excel.id_manager import IdManager
from usdm_excel.cdisc_ct_library import CDISCCTLibrary
from fhir.resources.bundle import Bundle, BundleEntry
from fhir.resources.composition import Composition, CompositionSection
from uuid import uuid4
from usdm_info import __model_version__ as usdm_version, __package_version__ as system_version

class FromFHIR():

  SYSTEM_NAME = "CDISC USDM FHIR"

  class LogicError(Exception):
    pass

  def __init__(self, errors_and_logging: ErrorsAndLogging):
    self._errors_and_logging = ErrorsAndLogging()
    self._id_manager = IdManager(self._errors_and_logging)
    self._cdisc_ct_manager = CDISCCTLibrary(self._errors_and_logging)

  def from_fhir(self, data: str) -> Wrapper:
    try:
      bundle = Bundle.parse_raw(data)
      study = self._study(bundle.entry[0].resource.title)
      doc_version = self._document_version(study)
      root = self._model_instance(NarrativeContent, {'name': 'ROOT', 'sectionNumber': '0', 'sectionTitle': 'Root', 'text': '', 'childIds': [], 'previousId': None, 'nextId': None})
      doc_version.contents.append(root)
      for item in bundle.entry[0].resource.section:
        nc_item = self._section(item, doc_version)
        root.childIds.append(nc_item.id)
      self._double_link(doc_version.contents, 'previousId', 'nextId')
      return Wrapper(study=study, usdmVersion=usdm_version, systemName=self.SYSTEM_NAME, systemVersion=system_version)
    except Exception as e:
      self._errors_and_logging.exception(f"Exception raised parsing FHIR content. See logs for more details", e)
      return None

  def _section(self, section: CompositionSection, doc_version):
    print(f"SECTION: {section.title}, {section.code.text}")
    nc = self._model_instance(NarrativeContent, {'name': f"{section.code.text}", 'sectionNumber': '', 'sectionTitle': section.title, 'text': section.text.div, 'childIds': [], 'previousId': None, 'nextId': None})
    doc_version.contents.append(nc)
    if section.section:
      for item in section.section:
        item_nc = self._section(item, doc_version)
        nc.childIds.append(item_nc.id)
    return nc
    
  def _study(self, title):
    sponsor_title_code = self._cdisc_ct_code('C99905x2', 'Official Study Title')
    protocl_status_code = self._cdisc_ct_code('C85255', 'Draft')
    intervention_model_code = self._cdisc_ct_code('C82639', 'Parallel Study')
    country_code = self._iso_country_code('DNK', 'Denmark')
    sponsor_code = self._cdisc_ct_code("C70793", 'Clinical Study Sponsor')
    study_title = self._model_instance(StudyTitle, {'text': title, 'type': sponsor_title_code})
    protocl_document_version = self._model_instance(StudyDefinitionDocumentVersion, {'protocolVersion': '1', 'protocolStatus': protocl_status_code})
    protocl_document = self._model_instance(StudyDefinitionDocument, {'name': 'PROTOCOL V1', 'label': '', 'description': '', 'versions': [protocl_document_version]})
    study_design = self._model_instance(StudyDesign, {'name': 'Study Design', 'label': '', 'description': '', 
      'rationale': 'XXX', 'interventionModel': intervention_model_code, 'arms': [], 'studyCells': [], 
      'epochs': [], 'population': None})
    address = self._model_instance(Address, {'line': 'Den Lille Havfrue', 'city': 'Copenhagen', 'district': '', 'state': '', 'postalCode': '12345', 'country': country_code})
    organization = self._model_instance(Organization, {'name': 'Sponsor', 'organizationType': sponsor_code, 'identifier': "123456789", 'identifierScheme': "DUNS", 'legalAddress': address}) 
    identifier = self._model_instance(StudyIdentifier, {'text': 'SPONSOR-1234', 'scope': organization})
    study_version = self._model_instance(StudyVersion, {'versionIdentifier': '1', 'rationale': 'XXX', 'titles': [study_title], 'studyDesigns': [study_design], 
                                                     'documentVersionId': protocl_document_version.id, 'studyIdentifiers': [identifier]}) 
    study = self._model_instance(Study, {'id': uuid4(), 'name': 'Study', 'label': '', 'description': '', 'versions': [study_version], 'documentedBy': protocl_document}) 
    return study

  def _cdisc_ct_code(self, code, decode):
    return self._model_instance(Code, {'code': code, 'decode': decode, 'codeSystem': self._cdisc_ct_manager.system, 'codeSystemVersion': self._cdisc_ct_manager.version})

  def _iso_country_code(self, code, decode):
    return self._model_instance(Code, {'code': code, 'decode': decode, 'codeSystem': 'ISO 3166 1 alpha3', 'codeSystemVersion': '2020-08'})
  
  def _document_version(self, study):
    return study.documentedBy.versions[0]
  
  def _model_instance(self, cls, params):
    params['id'] = params['id'] if 'id' in params else self._id_manager.build_id(cls)
    params['instanceType'] = cls.__name__
    return cls(**params)
  
  def _double_link(self, items, prev, next):
    for idx, item in enumerate(items):
      if idx == 0:
        setattr(item, prev, None)
      else:
        the_id = getattr(items[idx-1], 'id')
        setattr(item, prev, the_id)
      if idx == len(items)-1:  
        setattr(item, next, None)
      else:
        the_id = getattr(items[idx+1], 'id')
        setattr(item, next, the_id)


import re
from usdm_excel.globals import Globals
from usdm_excel.base_sheet import BaseSheet
from usdm_model.syntax_template_dictionary import SyntaxTemplateDictionary

class SyntaxTemplateSheet(BaseSheet):

  def __init__(self, file_path: str, globals: Globals, sheet_name: str, optional: bool=True):
    try:
      super().__init__(file_path=file_path, globals=globals, sheet_name=sheet_name, optional=optional)
    except Exception as e:
      self._sheet_exception(e)

  def _validate_references(self, row: int, column_name: str, text: str, dictionary_name: str) -> None:
    if dictionary_name:
      column = self.column_present(column_name)
      dictionary = self.globals.cross_references.get(SyntaxTemplateDictionary, dictionary_name)
      if not dictionary:
        self._warning(row, column, f"Dictionary '{dictionary_name}' not found")
        return
      tags = re.findall(r'\[([^]]*)\]',text)
      for tag in tags:
        entry = next((item for item in dictionary.parameterMaps if item.tag == tag), None)
        if not entry:
        #if not tag in dictionary.parameterMap:
          self._warning(row, column, f"Failed to find '{tag}' in dictionary '{dictionary_name}'")
  
  def _get_dictionary_id(self, dictionary_name: str) -> str:
    if dictionary_name:
      dictionary = self.globals.cross_references.get(SyntaxTemplateDictionary, dictionary_name)
      if dictionary:
        return dictionary.id
      else:
        self._general_error(f"Unable to find dictionary with name '{dictionary_name}'")
    return None



from src.usdm_excel.globals import Globals as GlobalsClass
from tests.test_factory import Factory as FactoryClass
from tests.test_data_factory import MinimalStudy
from usdm_model.eligibility_criterion import EligibilityCriterion
from usdm_excel.document.template_plain import TemplatePlain

def create_criteria(factory: FactoryClass, minimal: MinimalStudy):
  INCLUSION = factory.cdisc_code('C25532', 'Inc')
  EXCLUSION = factory.cdisc_code('C25370', 'Exc')
  item_list = [
    {'name': 'IE1', 'label': '', 'description': '', 'text': 'Only perform at baseline', 
     'dictionaryId': None, 'category': INCLUSION, 'identifier': '01', 'nextId': None, 'previousId': None, 'contextId': None
    },
    {'name': 'IE2', 'label': '', 'description': '', 'text': '<p>Only perform on males</p>', 
    'dictionaryId': None, 'category': INCLUSION, 'identifier': '02', 'nextId': None, 'previousId': None, 'contextId': None
    },
  ]
  results = factory.set(EligibilityCriterion, item_list)
  #print(f"RESULTS: {results}")
  for criterion in results:
    minimal.population.criterionIds.append(criterion.id)
  minimal.study.versions[0].criteria = results
  return results

def test_create(mocker, globals: GlobalsClass, minimal: MinimalStudy, factory: FactoryClass):
  globals.id_manager.clear()
  criteria = create_criteria(factory, minimal)
  bs = factory.base_sheet(mocker)
  template = TemplatePlain(bs, minimal.study_version, minimal.study_definition_document_version)
  result = template.inclusion({})
  expected = '<table class="table"><tr><td>01</td><td><usdm:ref klass="EligibilityCriterion" '\
    'id="EligibilityCriterion_1" attribute="text"></usdm:ref></td></tr><tr><td>02</td><td>'\
    '<usdm:ref klass="EligibilityCriterion" id="EligibilityCriterion_2" attribute="text"></usdm:ref>'\
    '</td></tr></table>'
  assert result == expected



import os
import datetime
import pandas as pd
from openpyxl import load_workbook
from usdm_excel.cdisc_ct import CDISCCT
from usdm_excel.other_ct import OtherCT
from usdm_excel.option_manager import Options
from usdm_excel.quantity_type import QuantityType
from usdm_excel.range_type import RangeType
from usdm_excel.iso_3166 import ISO3166
from usdm_model.quantity import Quantity
from usdm_model.range import Range
from usdm_model.address import Address
from usdm_model.comment_annotation import CommentAnnotation
from usdm_excel.alias import Alias
from usdm_excel.option_manager import EmptyNoneOption
from usdm_excel.globals import Globals
from usdm_model.geographic_scope import GeographicScope

class BaseSheet():

  class StateError(Exception):
    pass

  class FormatError(Exception):
    pass

  def __init__(self, file_path: str, globals: Globals, sheet_name: str, header: int=0, optional: bool=False, converters: dict={}, require: dict={}):
    self.file_path = file_path
    self.globals = globals
    self.dir_path, self.filename = os.path.split(file_path)
    self.sheet_name = sheet_name
    self.sheet = None
    self.success = False
    self._sheet_names = None
    if optional and not self._sheet_present(file_path, sheet_name):
      self._general_info(f"{sheet_name} not found but optional")
    else:
      if require and not self._check_cell_value(file_path, sheet_name, require['row'], require['column'], require['value']):
        self._general_info(f"Required value {require['value']} at [{require['row']}, {require['column']}] mismatch in {sheet_name}")
        pass
      else:
        self.sheet = pd.read_excel(open(file_path, 'rb'), sheet_name=sheet_name, header=header, converters=converters)
        self.success = True
        self._general_info("Processed sheet %s" % (sheet_name))

  def cell_empty(self, row_index, col_index):
    return pd.isnull(self.sheet.iloc[row_index, col_index])  

  def column_present(self, names):
    fields = [names] if isinstance(names, str) else names
    for field in fields:
      try:
        col_index = self.sheet.columns.get_loc(field)
        return col_index
      except:
        pass
    columns = ', '.join(fields)
    raise BaseSheet.FormatError(f"Failed to detect column(s) '{columns}' in sheet")

  def read_cell_by_name(self, row_index, field_name, default=None, must_be_present=True):
    try:
      col_index = self.column_present(field_name)
      return self.read_cell(row_index, col_index)
    except Exception as e:
      if not must_be_present:
        return "" #if self.globals.option_manager.get(Options.EMPTY_NONE) == EmptyNoneOption.EMPTY.value else None
      elif default:
        return default
      else:
        self._error(row_index, -2, f"Error attempting to read cell '{field_name}'. Exception: {e}")
        return "" #if self.globals.option_manager.get(Options.EMPTY_NONE) == EmptyNoneOption.EMPTY.value else None

  def read_cell(self, row_index, col_index, default=None):
    try:
      if pd.isnull(self.sheet.iloc[row_index, col_index]):
        if default:
          return default
        else:
          return "" #if self.globals.option_manager.get(Options.EMPTY_NONE) == EmptyNoneOption.EMPTY.value else None
      else:
        return str(self.sheet.iloc[row_index, col_index]).strip()
    except Exception as e:
      #print(f"ROW / COL: {row_index}, {col_index}")
      self._exception(row_index, col_index, 'Error reading cell', e)
      if default:
        return default
      else:
        return "" #if self.globals.option_manager.get(Options.EMPTY_NONE) == EmptyNoneOption.EMPTY.value else None

  # Deprecate this method
  def read_cell_empty_legacy(self, row_index, col_index):
    if self.cell_empty(row_index, col_index):
      return "", True
    else:
      value = self.read_cell_empty(row_index, col_index, '-')
      if value == "":
        return "", True
      else:
        return value, False

  # Deprecate this method
  def read_cell_empty(self, row_index, col_index, empty_character):
    value = self.read_cell(row_index, col_index)
    value = "" if (value == empty_character) or (value == None) else value
    return value

  def read_cell_multiple_by_name(self, row_index, field_name, must_be_present=True):
    try:
      col_index = self.column_present(field_name)
      return self.read_cell_multiple(row_index, col_index)
    except Exception as e:
      if not must_be_present:
        return []
      else:
        self._error(row_index, -2, f"Error '{e}' reading cell multiple '{field_name}'")
        return []

  def read_cell_multiple(self, rindex, cindex):
    try:
      results = []
      value = self.read_cell(rindex, cindex)
      if value.strip() == '':
        return results
      for part in self._state_split(value):
        results.append(part.strip())
      return results
    except BaseSheet.StateError as e:
      self._error(rindex, cindex, f"Internal state error '{e}' reading cell")
      return []
    except BaseSheet.FormatError as e:
      self._error(rindex, cindex, "Format error reading cell, check the format of the cell")
      return []
    
  def read_cell_with_previous(self, row_index, col_index, first_col_index):
    try:
      i = col_index
      while i >= first_col_index:
        if pd.isnull(self.sheet.iloc[row_index, i]):
          i -= 1
        else:
          return self.sheet.iloc[row_index, i].strip()
      self._warning(row_index, col_index, "No previous non-empty cell found.")
      return ""
    except Exception as e:
      self._error(row_index, col_index, "Error (%s) reading cell row '%s', field '%s'" % (e, row_index, col_index))
      return ""

  def read_boolean_cell_by_name(self, row_index, field_name, must_be_present=True):
    value = self.read_cell_by_name(row_index, field_name, must_be_present=must_be_present)
    if not value:
      return False
    elif value.strip().upper() in ['Y', 'YES', 'T', 'TRUE', '1']:
      return True
    return False

  def read_date_cell_by_name(self, row_index, field_name, must_be_present=True):
    col_index = self.column_present(field_name)
    return self.read_date_cell(row_index, col_index, must_be_present)

  def read_date_cell(self, row_index, col_index, must_be_present=True):
    cell = self.read_cell(row_index, col_index)
    try:
      print(f"DATE: {cell}")
      return datetime.datetime.strptime(cell, '%Y-%m-%d %H:%M:%S')
    except Exception as e:
      self._exception(row_index, col_index, "Error reading date cell", e)
      return None

  def read_quantity_cell_by_name(self, row_index, field_name, allow_missing_units=True, allow_empty=True):
    col_index = self.column_present(field_name)
    return self.read_quantity_cell(row_index, col_index, allow_missing_units, allow_empty)

  def read_quantity_cell(self, row_index, col_index, allow_missing_units=True, allow_empty=True):
    try:
      text = self.read_cell(row_index, col_index)
      quantity = QuantityType(text, self.globals, allow_missing_units, allow_empty)
      if not quantity.errors:
        unit = Alias(self.globals).code(quantity.units_code, [])
        return None if quantity.empty else Quantity(id=self.globals.id_manager.build_id(Quantity), value=float(quantity.value), unit=unit)
      else:
        self._add_errors(quantity.errors, row_index, col_index)
        return None
    except Exception as e:
      self._error( row_index, col_index, f"Failed to decode quantity data '{text}'", e)
      return None

  def read_range_cell_by_name(self, row_index, field_name, require_units=True, allow_empty=True):
    col_index = self.column_present(field_name)
    return self.read_range_cell(row_index, col_index, require_units, allow_empty)

  def read_range_cell(self, row_index, col_index, require_units=True, allow_empty=True):
    try:
      text = self.read_cell(row_index, col_index)
      range = RangeType(text, self.globals, require_units, allow_empty)
      if not range.errors:
        #print(f"RANGE: {range.lower} {range.upper} {range.units} {range.units_code} {range.empty} ")
        return None if range.empty else Range(id=self.globals.id_manager.build_id(Range), minValue=float(range.lower), maxValue=float(range.upper), unit=range.units_code, isApproximate=False)
      else:
        self._add_errors(range.errors, row_index, col_index)
        return None
    except Exception as e:
      self._error( row_index, col_index, f"Failed to decode quantity data '{text}'", e)
      return None

  def read_address_cell_by_name(self, row_index, field_name, allow_empty=False):
    raw_address = self.read_cell_by_name(row_index, field_name)
    # TODO The '|' separator is preserved for legacy reasons but should be removed in the future
    if not raw_address:
      sep = ','
      parts = []
    elif '|' in raw_address:
      sep = '|'
      parts = raw_address.split(sep)
    else:
      sep = ','
      parts = self._state_split(raw_address)
    if len(parts) >= 6:
      result = self._to_address(
          self.globals.id_manager.build_id(Address),
          lines=[x.strip() for x in parts[:-5]], 
          district=parts[-5].strip(), 
          city=parts[-4].strip(), 
          state=parts[-3].strip(), 
          postal_code=parts[-2].strip(), 
          country=ISO3166(self.globals).code(parts[-1].strip())
        )
      return result
    elif allow_empty:
      return None
    else:
      col_index = self.column_present(field_name)
      self._error(row_index, col_index, f"Address '{raw_address}' does not contain the required fields (lines, district, city, state, postal code and country code) using '{sep}' separator characters, only {len(parts)} found")
      return None

  def read_geographic_scopes_cell_by_name(self, row_index, field_name):
    try:
      col_index = self._get_column_index(field_name)
      return self.read_geographic_scopes_cell(row_index, col_index)
    except Exception as e:
      self._warning(row_index, -2, "No geographic scope column found, assuming global scope.")
      return [self._scope('Global', None)]
  
  def read_geographic_scopes_cell(self, row_index, col_index):
    result = []
    value = self.read_cell(row_index, col_index, default='')
    if value.strip() == '':
      self._warning(row_index, col_index, "Empty cell detected where geographic scope values expected, assuming global scope.")
      result.append(self._scope('Global', None))
    else:
      for item in self._state_split(value):
        key_value = self._key_value(item, row_index, col_index, allow_single=True)
        if key_value[0] == "GLOBAL":
          result.append(self._scope('Global', None))
        elif key_value[0] == "REGION": 
          code = self._country_region(key_value[1], 'Region')
          if code:
            scope = self._scope('Region', code)
            result.append(scope)
        elif key_value[0] == "COUNTRY": 
          code = self._country_region(key_value[1], 'Country')
          if code:
            scope = self._scope('Country', code)
            result.append(scope)
        else:
          self._warning(row_index, col_index, f"Failed to decode geographic scope '{value}'. Formats are 'Global', 'Region: <value>' or 'Country: <value>'. Assuming global scope.")
          result.append(self._scope('Global', None))
    return result

  def _key_value(self, text: str, row_index: int, col_index: int, allow_single=False):
    if text.strip():
      parts = text.split(":")
      if len(parts) == 2:
        return [parts[0].strip().upper(), parts[1].strip()]
      elif len(parts) == 1 and allow_single:
        return [parts[0].strip().upper(), '']
    self._error(row_index, col_index, f"Failed to decode key value pair '{text}', incorrect format, missing ':'?")
    return ['', '']

  def _country_region_quantity(self, text: str, type: str, row_index: int, col_index: int):
    name_value = text.split('=')
    if len(name_value) == 2:
      quantity = self._get_quantity(name_value[1].strip())
      code = self._country_region(name_value[0].strip(), type)
      return code, quantity
    else:
      self._error(row_index, col_index, f"Failed to decode geographic enrollment data '{text}', incorrect format, missing '='?")
      return None, None

  def _country_region(self, text: str, type: str):
    return ISO3166(self.globals).region_code(text) if type == 'Region' else ISO3166(self.globals).code(text)

  def _get_quantity(self, text):
    quantity = QuantityType(text, self.globals, True, False)
    unit = Alias(self.globals).code(quantity.units_code, [])
    return self.create_object(Quantity, {'value': float(quantity.value), 'unit': unit})

  def _scope(self, type, code):
    scope_type = CDISCCT(self.globals).code_for_attribute('GeographicScope', 'type', type)
    alias = Alias(self.globals).code(code, []) if code else None
    return self.create_object(GeographicScope, {'type': scope_type, 'code': alias})

  def _to_address(self, id, lines, city, district, state, postal_code, country):
    text = f"{(', ').join(lines)}, {city}, {district}, {state}, {postal_code}, {country.decode}"
    #text = text.replace(' ,', '')
    try:
      result = Address(id=id, text=text, lines=lines, city=city, district=district, state=state, postalCode=postal_code, country=country)
    except Exception as e:
      self._general_exception(f"Failed to create Address object", e)
      result = None
    return result

  def create_object(self, cls, params):
    try:
      params['id'] = self.globals.id_manager.build_id(cls)
      return cls(**params)
    except Exception as e:
      self._general_exception(f"Failed to create {cls.__name__} object", e)
      return None

  def add_notes(self, instance: object, note_refs: list) -> None:
    for note_ref in note_refs:
      try:
        note = self.globals.cross_references.get(CommentAnnotation, note_ref)
        if note:
          instance.notes.append(note)
        else:
          self._general_error(f"Failed to find note with name '{note_ref}'")  
      except Exception as e:
        self._general_exception(f"Failed to add note to '{object.__class__.__name__}' object", e)

  def read_other_code_cell_by_name(self, row_index, field_name):
    col_index = self.column_present(field_name)
    return self.read_other_code_cell(row_index, col_index)

  def read_other_code_cell(self, row_index, col_index):
    value = self.read_cell(row_index, col_index)
    if value.strip() == "":
      return None
    return self._decode_other_code(value, row_index, col_index)

  def read_other_code_cell_multiple_by_name(self, row_index, field_name):
    col_index = self.column_present(field_name)
    return self.read_other_code_cell_mutiple(row_index, col_index)

  def read_other_code_cell_mutiple(self, row_index, col_index):
    value = self.read_cell(row_index, col_index)
    result = []
    if value.strip() == '':
      return result
    for item in self._state_split(value):
      code = self._decode_other_code(item.strip(), row_index, col_index)
      if not code == None:
        result.append(code)
    return result

  def read_cdisc_klass_attribute_cell_by_name(self, klass, attribute, row_index, field_name, allow_empty=False):
    col_index = self.column_present(field_name)
    return self.read_cdisc_klass_attribute_cell(klass, attribute, row_index, col_index, allow_empty)

  def read_cdisc_klass_attribute_cell(self, klass, attribute, row_index, col_index, allow_empty=False):
    code = None
    value = self.read_cell(row_index, col_index)
    if value:
      code = CDISCCT(self.globals).code_for_attribute(klass, attribute, value)
      if not code:
        self._error(row_index, col_index, f"CDISC CT not found for value '{value}'.")
    elif not allow_empty:
      self._error(row_index, col_index, "Empty cell detected where CDISC CT value expected.")
    return code

  def read_cdisc_klass_attribute_cell_multiple_by_name(self, klass, attribute, row_index, field_name):
    col_index = self.column_present(field_name)
    return self.read_cdisc_klass_attribute_cell_multiple(klass, attribute, row_index, col_index)

  def read_cdisc_klass_attribute_cell_multiple(self, klass, attribute, row_index, col_index):
    result = []
    value = self.read_cell(row_index, col_index)
    if value.strip() == "":
      self._error(row_index, col_index, "Empty cell detected where multiple CDISC CT values expected.")
      return result
    for item in self._state_split(value):
      code = CDISCCT(self.globals).code_for_attribute(klass, attribute, item.strip())
      if code is not None:
        result.append(code)
      else:
        self._error(row_index, col_index, f"CDISC CT not found for value '{item.strip()}'.")
    return result

  def _get_cross_reference(self, klass, name, error_klass_name):
    item = self.globals.cross_references.get(klass, name)
    if item:
      return item.id
    else:
      self._general_error(f"Unable to find {error_klass_name} with name '{name}'")              
      return None

  def double_link(self, items, prev, next):
    try: 
      for idx, item in enumerate(items):
        if idx == 0:
          if self.globals.option_manager.get(Options.EMPTY_NONE) == EmptyNoneOption.EMPTY.value:
            setattr(item, prev, "")
          else:
            setattr(item, prev, None)
        else:
          the_id = getattr(items[idx-1], 'id')
          setattr(item, prev, the_id)
        if idx == len(items)-1:  
          if self.globals.option_manager.get(Options.EMPTY_NONE) == EmptyNoneOption.EMPTY.value:
            setattr(item, next, "")
          else:
            setattr(item, next, None)
        else:
          the_id = getattr(items[idx+1], 'id')
          setattr(item, next, the_id)
    except Exception as e:
      self._general_exception(f"Error while doubly linking lists", e)

  def previous_link(self, items, prev):
    try: 
      for idx, item in enumerate(items):
        if idx == 0:
          if self.globals.option_manager.get(Options.EMPTY_NONE) == EmptyNoneOption.EMPTY.value:
            setattr(item, prev, "")
          else:
            setattr(item, prev, None)
        else:
          the_id = getattr(items[idx-1], 'id')
          setattr(item, prev, the_id)
    except Exception as e:
      self._general_exception(f"Error in previous link link {items}", e)

  def _decode_other_code(self, value, row_index, col_index):
    if value.strip() == "":
      return None
    outer_parts = value.split(":")
    if len(outer_parts) == 2:
      system = outer_parts[0].strip()
      inner_parts = outer_parts[1].strip().split("=")
      if len(inner_parts) == 2:
        version = self.globals.ct_version_manager.get(system)
        return OtherCT(self.globals).code(code=inner_parts[0].strip(), system=system, version=version, decode=inner_parts[1].strip())
      else:
        self._error(row_index, col_index, "Failed to decode code data '%s', no '=' detected" % (value))
    else:
      self._error(row_index, col_index, "Failed to decode code data '%s', no ':' detected" % (value))
    return None

  def _to_int(self, value):
    try:
      return int(value)
    except:
      return None

  def _get_column_index(self, column_name):
    try:
      col_index = self.sheet.columns.get_loc(column_name)
      return col_index
    except Exception as e:
      pass
    raise BaseSheet.FormatError(f"Failed to detect column(s) '{column_name}' in sheet")

  def _add_errors(self, errors, row, column):
    for error in errors:
      self._error(row, column, error)
      
  def _info(self, row: int, column: int | str, message: str):
    try:
      column = self._get_column_index(column) if isinstance(column, str) else column
      self.globals.errors_and_logging.info(message, self.sheet_name, row + 1, column + 1)
    except Exception as e:
      self.globals.errors_and_logging.exception(message, e, self.sheet_name)
     
  def _general_info(self, message):
    self.globals.errors_and_logging.info(message, self.sheet_name)
     
  def _error(self, row: int, column: int | str, message: str):
    try:
      column = self._get_column_index(column) if isinstance(column, str) else column
      self.globals.errors_and_logging.error(message, self.sheet_name, row + 1, column + 1)
    except Exception as e:
      self.globals.errors_and_logging.exception(message, e, self.sheet_name)

  def _general_error(self, message):
    self.globals.errors_and_logging.error(message, self.sheet_name)

  def _warning(self, row: int, column: int | str, message: str):
    try:
      column = self._get_column_index(column) if isinstance(column, str) else column
      self.globals.errors_and_logging.warning(message, self.sheet_name, row + 1, column + 1)
    except Exception as e:
      self.globals.errors_and_logging.exception(message, e, self.sheet_name)

  def _general_warning(self, message):
    self.globals.errors_and_logging.warning(message, self.sheet_name)

  def _debug(self, row: int, column: int | str, message: str):
    try:
      column = self._get_column_index(column) if isinstance(column, str) else column
      self.globals.errors_and_logging.debug(message, self.sheet_name, row + 1, column + 1)
    except Exception as e:
      self.globals.errors_and_logging.exception(message, e, self.sheet_name)

  def _general_debug(self, message):
    self.globals.errors_and_logging.debug(message, self.sheet_name)

  def _general_exception(self, message, e):
    self.globals.errors_and_logging.exception(message, e, self.sheet_name)

  def _exception(self, row: int, column: int | str, message: str, e: Exception):
    try:
      column = self._get_column_index(column) if isinstance(column, str) else column
      self.globals.errors_and_logging.exception(message, e, self.sheet_name, row + 1, column + 1)
    except Exception as e:
      self.globals.errors_and_logging.exception(message, e, self.sheet_name)

  def _sheet_exception(self, e):
    self.globals.errors_and_logging.exception(f"Error [{e}] while reading sheet '{self.sheet_name}'", e, self.sheet_name)

  def _get_sheet_names(self, file_path):
    if not self._sheet_names:
      wb = load_workbook(file_path, read_only=True, keep_links=False)
      self._sheet_names = wb.sheetnames
    return self._sheet_names
  
  def _sheet_present(self, file_path, sheet_name):
    sheet_names = self._get_sheet_names(file_path)
    return sheet_name in sheet_names

  def _check_cell_value(self, file_path, sheet_name, row, column, value):
    wb = load_workbook(file_path, read_only=True, keep_links=False)
    ws = wb[sheet_name]
    #print(f"CELL={ws.cell(row, column).value}")
    return str(ws.cell(row, column).value).upper() == value

  def _state_split(self, s):

    OUT = "out"
    IN_QUOTED = "in_quoted"
    OUT_QUOTED = "out_quoted"
    IN_NORMAL = "in_normal"
    ESC = "escape"

    state = OUT
    result = []
    current = []
    exit = ""
    for c in s:
      #print(f"STATE: s: {state}, c: {c}")
      if state == OUT:
        current = []
        if c == ',':
          result.append("")
        elif c in ['"', "'"]:
          state = IN_QUOTED
          exit = c
        elif c.isspace():
          pass
        else:
          state = IN_NORMAL
          current.append(c)
      elif state == IN_QUOTED:
        if c == '\\':
          state = ESC
        elif c == exit:
          result.append("".join(current).strip())
          state = OUT_QUOTED
        else:
          current.append(c)
      elif state == OUT_QUOTED:
        if c == ',':
          state = OUT
        else:
          pass
      elif state == IN_NORMAL:
        if c == ',':
          result.append("".join(current).strip())
          state = OUT
        else:
          current.append(c)
      elif state == ESC:
        if c == exit:
          current.append(c)
          state = IN_QUOTED
        else:
          current.append('\\')
          current.append(c)
      else:
        raise BaseSheet.StateError

    if state == OUT or state == OUT_QUOTED:
      pass
    elif state == IN_NORMAL:
      result.append("".join(current).strip())
    else:
      raise BaseSheet.FormatError
    return result



import os
import yaml
import traceback
import requests
from usdm_excel.errors_and_logging.errors_and_logging import ErrorsAndLogging

class CDISCCTLibrary():

  API_ROOT = 'https://api.library.cdisc.org/api'  
  HEADERS = { "Content-Type":"application/json", "api-key": os.getenv('CDISC_API_KEY') }
    
  def __init__(self, errors_and_logging: ErrorsAndLogging):
    self._errors_and_logging = errors_and_logging
    f = open(os.path.join(os.path.dirname(__file__), 'data', 'missing_ct.yaml'))
    self._missing_ct = yaml.load(f, Loader=yaml.FullLoader)
    f = open(os.path.join(os.path.dirname(__file__), 'data', 'cdisc_ct_config.yaml'))
    self._cdisc_ct_config = yaml.load(f, Loader=yaml.FullLoader)
    # Visible properties
    self.version = self._cdisc_ct_config['version']
    self.system = "http://www.cdisc.org"
    # Private properties
    self._by_code_list = {}
    self._by_term = {}
    self._by_submission = {}
    self._by_pt = {}
    self._by_klass_attribute = {}
    if self._code_lists_exist():
      self._load_ct()
    else:
      self._get_ct()
      self._save_code_lists(self._by_code_list)
    self._get_missing_ct()
    self._get_klass_attribute()

  def submission(self, value, cl=None):
    if value in list(self._by_submission.keys()):
      concept_ids = self._by_submission[value]
      if len(concept_ids) == 0:
        return None
      elif len(concept_ids) == 1:
        code_list = self._by_code_list[concept_ids[0]]
        return next((item for item in code_list['terms'] if item["submissionValue"] == value), None)
      else:
        if cl and cl in concept_ids:
          code_list = self._by_code_list[cl]
          return next((item for item in code_list['terms'] if item["submissionValue"] == value), None)
        else:
          return None 
    else:
      return None
    
  def preferred_term(self, value, cl=None):
    if value in list(self._by_pt.keys()):
      concept_ids = self._by_pt[value]
      if len(concept_ids) == 0:
        return None
      elif len(concept_ids) == 1:
        code_list = self._by_code_list[concept_ids[0]]
        return next((item for item in code_list['terms'] if item["preferredTerm"] == value), None)
      else:
        if cl and cl in concept_ids:
          code_list = self._by_code_list[cl]
          return next((item for item in code_list['terms'] if item["preferredTerm"] == value), None)
        else:
          return None 
    else:
      return None

  def klass_and_attribute(self, klass, attribute, value):
    try:
      #if value == "Protocol Treatment Arm":
      #  print(f"Klass: {klass}, attribute: {attribute}, value: {value}")
      concept_id = self._by_klass_attribute[klass][attribute]
      #if value == "Protocol Treatment Arm":
      #  print(f"Concept: {concept_id}")
      code_list = self._by_code_list[concept_id]
      #if value == "Protocol Treatment Arm":
      #  print(f"Concept: {code_list}")
      return self._get_item(code_list, value)
    except Exception as e: 
      self._errors_and_logging.exception(f"Failed to find '{value}' for klass '{klass}' attribute '{attribute}'", e)
      return None

  def unit(self, value):
    try:
      code_list = self._by_code_list['C71620']
      return self._get_item(code_list, value)
    except Exception as e: 
      self._errors_and_logging.exception(f"Failed to find unit '{value}'") 
      return None

  def _get_item(self, code_list, value):
    try:
      for field in [ 'conceptId', 'preferredTerm', 'submissionValue']:
        #if value == "Protocol Treatment Arm":
        #  print(f"GET: {field}")
        result = next((item for item in code_list['terms'] if item[field].upper() == value.upper()), None)
        if result:
          return result
      return None
    except Exception as e:
      self._errors_and_logging.exception(f"Failed to find CDSIC CT for '{value}' in code ist '{code_list}'", e)
      return None

  def _get_ct(self):
    for item in self._cdisc_ct_config['required']:
      self._get_code_list(item)

  def _load_ct(self):
    self._by_code_list = self._read_code_lists()
    for c_code, entry in self._by_code_list.items():
      for item in entry['terms']:
        self._check_in_and_add(self._by_term, item['conceptId'], c_code)
        self._check_in_and_add(self._by_submission, item['submissionValue'], c_code)
        self._check_in_and_add(self._by_pt, item['preferredTerm'], c_code)
 
  def _get_missing_ct(self):
    for response in self._missing_ct:
      self._by_code_list[response['conceptId']] = response
      for item in response['terms']:
        self._check_in_and_add(self._by_term, item['conceptId'], response['conceptId'])
        self._check_in_and_add(self._by_submission, item['submissionValue'], response['conceptId'])
        self._check_in_and_add(self._by_pt, item['preferredTerm'], response['conceptId'])

  def _get_klass_attribute(self):
    for klass, info in self._cdisc_ct_config['klass'].items():
      if not klass in self._by_klass_attribute:
        self._by_klass_attribute[klass] = {}
      for attribute, cl in info.items():
        if not attribute in self._by_klass_attribute[klass]:
          self._by_klass_attribute[klass][attribute] = cl

  def _get_code_list(self, c_code):
    for package in self._cdisc_ct_config['packages']:
      package_full_name = "%sct-%s" % (package, self.version)
      api_url = self._url('/mdr/ct/packages/%s/codelists/%s' % (package_full_name, c_code))
      self._errors_and_logging.info(f"CDISC CT Library: {api_url}")
      raw = requests.get(api_url, headers=self.__class__.HEADERS)
      if raw.status_code == 200:
        response = raw.json()
        response.pop('_links', None)
        self._by_code_list[response['conceptId']] = response
        for item in response['terms']:
          self._check_in_and_add(self._by_term, item['conceptId'], response['conceptId'])
          self._check_in_and_add(self._by_submission, item['submissionValue'], response['conceptId'])
          self._check_in_and_add(self._by_pt, item['preferredTerm'], response['conceptId'])
        return
    # If none found in all of the packages, then log error.
    self._errors_and_logging.exception(f"Failed to find CDSIC CT for C code '{c_code}'")

  def _url(self, relative_url):
    return f"{self.__class__.API_ROOT}{relative_url}"

  def _check_in_and_add(self, collection, id, item):
    if not id in collection:
      collection[id] = []
    collection[id].append(item)

  def _save_code_lists(self, data):
    try:
      if not self._code_lists_exist():
        with open(self._ct_filename(), 'w') as f:
          yaml.dump(data, f, indent=2, sort_keys=True)
    except Exception as e:
      self._errors_and_logging.exception("Failed to save CDSIC CT file", e)

  def _read_code_lists(self):
    try:
      if self._code_lists_exist():
        with open(self._ct_filename()) as f:
          return yaml.safe_load(f)
      else:
        self._errors_and_logging.error(f"Failed to read CDSIC CT file, does not exist")
        return None
    except Exception as e:
      self._errors_and_logging.exception("Failed to read CDSIC CT file", e)

  def _code_lists_exist(self):
    return os.path.isfile(self._ct_filename()) 

  def _ct_filename(self):
    return os.path.join(os.path.dirname(__file__), 'data', f"cdisc_ct_{self._cdisc_ct_config['version']}.yaml")



from usdm_excel.globals import Globals
from usdm_excel.cdisc_ct_library import CDISCCTLibrary
from usdm_excel.code_base import CodeBase

class CDISCCT(CodeBase):

  def __init__(self, globals: Globals):
    super().__init__(globals)
    self._library = self._globals.cdisc_ct_library
    
  def code(self, code, decode):
    return self._build(code=code, system=self._library.system, version=self._library.version, decode=decode)
 
  def code_for_attribute(self, klass, attribute, value):
    item = self._library.klass_and_attribute(klass, attribute, value)
    if item:
      return self._build(code=item['conceptId'], system=self._library.system, version=self._library.version, decode=item['preferredTerm'])
    else:
      return None

  def code_for_unit(self, value):
    item = self._library.unit(value)
    if item == None:
      return None
    return self._build(code=item['conceptId'], system=self._library.system, version=self._library.version, decode=item['preferredTerm'])



from usdm_excel.base_sheet import BaseSheet
from usdm_model.narrative_content import NarrativeContent, NarrativeContentItem
from usdm_excel.globals import Globals

class DocumentTemplates():

  def __init__(self, file_path: str, globals: Globals):
    self.items = []
    for template, sheet in globals.template_manager.items():
      document = DocumentTemplateSheet(file_path, template, sheet, globals)
      self.items.append(document)

class DocumentTemplateSheet(BaseSheet):

  def __init__(self, file_path: str, template_name: str, sheet_name: str, globals: Globals):
    try:
      self.items = []
      self.name = template_name
      super().__init__(file_path=file_path, globals=globals, sheet_name=sheet_name, optional=True, converters={"sectionName": str})
      if self.success:
        current_level = 0
        new_level = 0
        self._parent_stack = []
        previous_item = None
        for index, row in self.sheet.iterrows():
          name = self.read_cell_by_name(index, 'name')
          section_number = self.read_cell_by_name(index, 'sectionNumber')
          name = f"SECTION {section_number}" if not name else name
          new_level = self._get_level(section_number)
          section_title = self.read_cell_by_name(index, 'sectionTitle')
          display_section_number = self.read_boolean_cell_by_name(index, 'displaySectionNumber')
          display_section_title = self.read_boolean_cell_by_name(index, 'displaySectionTitle')
          content_name = self.read_cell_by_name(index, 'content')
          #print(f"TEMPLATE: '{content_name}'")
          if content_name:
            content = self.globals.cross_references.get(NarrativeContentItem, content_name)
            if not content:
              self._general_warning(f"Unable to find content item with name '{content_name}'")
          else:
            self._general_warning(f"No content item specified for section '{section_number}', '{section_title}'")
            content = None
          params = {
            'name': name, 
            'sectionNumber': section_number, 
            'displaySectionNumber': display_section_number, 
            'sectionTitle': section_title, 
            'displaySectionTitle': display_section_title, 
            'contentItemId': content.id if content else None
          }
          item = self.create_object(NarrativeContent, params)
          if item:
            self.items.append(item)
            self.globals.cross_references.add(name, item)     
            if new_level == current_level:
              # Same level
              self._add_child_to_parent(item)
            elif new_level > current_level:
              # Down
              if (new_level - current_level) > 1:
                self._error(index, self._get_column_index('sectionNumber'), f"Error with section number incresing by more than one level, section '{section_number}'.")
                raise BaseSheet.FormatError
              if previous_item:
                self._push_parent(previous_item)
              self._add_child_to_parent(item)
              current_level = new_level
            else:
              # Up
              self._pop_parent(current_level, new_level)
              self._add_child_to_parent(item)
              current_level = new_level
            previous_item = item
          self.double_link(self.items, 'previousId', 'nextId')
    except Exception as e:
      self._sheet_exception(e)

  def _get_level(self, section_number):
    sn = section_number[:-1] if section_number.endswith('.') else section_number
    parts = sn.split('.')
    return len(parts)

  def _push_parent(self, parent):
    self._parent_stack.append(parent)

  def _pop_parent(self, current_level, new_level):
    for p_count in range(new_level, current_level):
      popped = self._parent_stack.pop()
  
  def _add_child_to_parent(self, child):
    if self._parent_stack:
      parent = self._parent_stack[-1]
      parent.childIds.append(child.id)


import pytest
import pandas as pd
from usdm_excel.study_design_characteristics_sheet.study_design_characteristics_sheet import StudyDesignCharacteristicSheet

xfail = pytest.mark.xfail
def test_create(mocker, globals):
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['CharacteristicId_1', 'CharacteristicId_2', 'CharacteristicId_3']
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['CHAR01', 'The study age characteristic', 'Age characteristic', 'Subjects should be between 18 and 45 years old', "dictionary"], 
    ['CHAR02', 'The study abc characteristic',   'ABC characteristic', 'Subjects should have ABC', "dictionary"], 
    ['CHAR03', 'Exclude those with all fingers',   'Fingers characteristic', 'Subjects should not have all fingers', "dictionary"], 
 ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['name', 'description', 'label', 'text', 'dictionary'])
  items = StudyDesignCharacteristicSheet("", globals)
  assert len(items.items) == 3
  assert items.items[0].id == 'CharacteristicId_1'
  assert items.items[0].name == 'CHAR01'
  assert items.items[0].description == 'The study age characteristic'
  assert items.items[0].label == 'Age characteristic'
  assert items.items[0].text == 'Subjects should be between 18 and 45 years old'
  
def test_create_empty(mocker, globals):
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['name', 'description', 'label', 'text', 'dictionary'])
  items = StudyDesignCharacteristicSheet("", globals)
  assert len(items.items) == 0

def test_read_cell_by_name_error(mocker, globals):
  globals.cross_references.clear()
  call_parameters = []
  
  def my_add(sheet, row, column, message, level=10):
    call_parameters.append((sheet, row, column, message, level))
    return None

  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add", side_effect=my_add)
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['The study age characteristic', 'Age characteristic', 'Subjects should be between 18 and 45 years old', "dictionary"]
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['name', 'description', 'label', 'dictionary'])
  items = StudyDesignCharacteristicSheet("", globals)
  mock_error.assert_called()
  assert call_parameters == [
    ('studyDesignCharacteristics', 1, -1, "Error attempting to read cell 'text'. Exception: Failed to detect column(s) 'text' in sheet", 40),
    ('studyDesignCharacteristics', None, None, "Exception. Error [Failed to detect column(s) 'text' in sheet] while reading sheet 'studyDesignCharacteristics'. See log for additional details.", 40),
  ]
  



import traceback
from usdm_excel.base_sheet import BaseSheet
from usdm_model.study_arm import StudyArm
from usdm_excel.globals import Globals

class StudyDesignArmSheet(BaseSheet):

  SHEET_NAME = 'studyDesignArms'
  
  def __init__(self, file_path: str, globals: Globals):
    try:
      self.items = []
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME)
      for index, row in self.sheet.iterrows():
        name = self.read_cell_by_name(index, ['studyArmName', 'name'])
        description = self.read_cell_by_name(index, ['studyArmDescription', 'description'])
        label = self.read_cell_by_name(index, 'label', default="", must_be_present=False)
        arm_type = self.read_cdisc_klass_attribute_cell_by_name('StudyArm', 'studyArmType', index, ['studyArmType', 'type'])
        arm_origin_description = self.read_cell_by_name(index, ['studyArmDataOriginDescription', 'dataOriginDescription'])
        arm_origin_type = self.read_cdisc_klass_attribute_cell_by_name('StudyArm', 'studyArmDataOriginType', index, ['studyArmDataOriginType', 'dataOriginType'])
        notes = self.read_cell_multiple_by_name(index, 'notes', must_be_present=False)
        item = self.create_object(StudyArm, {'name': name, 'description': description, 'label': label, 'type': arm_type, 'dataOriginDescription': arm_origin_description, 'dataOriginType': arm_origin_type})
        if item:
          self.items.append(item)
          self.globals.cross_references.add(name, item)
          self.add_notes(item, notes)
    except Exception as e:
      self._sheet_exception(e)




import pandas as pd
from usdm_excel.study_design_intervention_sheet.study_design_product_sheet import StudyDesignProductSheet
from usdm_excel.base_sheet import BaseSheet
from usdm_model.code import Code

def test_create_1(mocker, globals):
  ids = ['Id_1', 'Id_2', 'Id_3', 'Id_4', 'Id_5', 'Id_6', 'Id_7', 'Id_8', 'Id_9', 'Id_10', 'Id_11', 'Id_12', 'Id_13', 'Id_14']
  expected = ( 
    '{"id": "Id_4", "name": "60 mg Study Drug", "label": "label 1", "description": "description 1", "pharmacologicClass": '
      '{"id": "Id_3", "code": "A", "codeSystem": "FDA", "codeSystemVersion": "", "decode": "B", "instanceType": "Code"}, '
    '"administrableDoseForm": '
      '{"id": "Id_2", '
        '"standardCode": {"id": "Id_1", "code": "C42998", "codeSystem": "http://www.cdisc.org", "codeSystemVersion": "2023-12-15", "decode": "Tablet Dosage Form", "instanceType": "Code"}, '
        '"standardCodeAliases": [], '
        '"instanceType": "AliasCode"}, '
      '"properties": [], "identifiers": [], '
      '"ingredients": '
        '[{"id": "Id_7", "role": '
          '{"id": "Id_6", "code": "100000072072", "codeSystem": "HL7", "codeSystemVersion": "", "decode": "Active", "instanceType": "Code"}, '
        '"substance": '
          '{"id": "Id_5", "name": "Ingredient C", "label": "label 2", "description": "description 2", "codes": [], '
          '"strengths": '
            '[{"id": "Id_14", "name": "60 mg", "label": "", "description": "", '
            '"denominator": '
              '{"id": "Id_13", "value": 1.0, '
              '"unit": '
                '{"id": "Id_12", '
                '"standardCode": '
                  '{"id": "Id_11", "code": "C48542", "codeSystem": "http://www.cdisc.org", "codeSystemVersion": "2023-12-15", "decode": "Tablet Dosing Unit", "instanceType": "Code"}, '
                '"standardCodeAliases": [], "instanceType": "AliasCode"}, '
              '"instanceType": "Quantity"}, '
            '"numerator": {"id": "Id_10", "value": 60.0, '
              '"unit": {"id": "Id_9", '
                '"standardCode": {"id": "Id_8", "code": "C28253", "codeSystem": "http://www.cdisc.org", "codeSystemVersion": "2023-12-15", "decode": "Milligram", "instanceType": "Code"}, '
                '"standardCodeAliases": [], "instanceType": "AliasCode"}, '
              '"instanceType": "Quantity"}, '
            '"instanceType": "Strength"}], '
          '"referenceSubstance": null, "instanceType": "Substance"}, '
        '"instanceType": "Ingredient"}], "notes": [], '
      '"instanceType": "AdministrableProduct"}'
  )
  data = {
    'name': ['60 mg Study Drug'],
    'description': ['description 1'],
    'label': ['label 1'],
    'pharmacologicClass': ['FDA: A=B'],
    'administrableDoseForm': ['TABLET'],
    'ingredientRole': ['HL7:   100000072072=Active'],
    'substanceName': ['Ingredient C'],
    'substanceDescription': ['description 2'],
    'substanceLabel': ['label 2'],
    'substanceCode': [''],
    'strengthName': ['60 mg'],
    'strengthDescription': [''],
    'strengthLabel': [''],
    'strengthNumerator': ['60 mg'],
    'strengthDenominator': ['1 TABLET'],
    'referenceSubstanceName': [''],
    'referenceSubstanceDescription': [''],
    'referenceSubstanceLabel': [''],
    'referenceSubstanceCode': [''],
    'referenceSubstanceStrengthName': [''],
    'referenceSubstanceStrengthDescription': [''],
    'referenceSubstanceStrengthLabel': [''],
    'referenceSubstanceStrengthNumerator': [''],
    'referenceSubstanceStrengthDenominator': ['']
  }  
  sheet = _setup_sheet(mocker, globals, data, ids)
  assert str(sheet.items[0].to_json()) == expected

def test_create_2(mocker, globals):
  ids = [
     'Id_1',  'Id_2',  'Id_3',  'Id_4',  'Id_5',  'Id_6',  'Id_7',  'Id_8',  'Id_9', 'Id_10', 
    'Id_11', 'Id_12', 'Id_13', 'Id_14', 'Id_15', 'Id_16', 'Id_17', 'Id_18', 'Id_19', 'Id_20',
    'Id_21', 'Id_22', 'Id_23', 'Id_24', 'Id_25', 'Id_26', 'Id_27', 'Id_28', 'Id_29', 'Id_30'
  ]
  data = {
    'name': ['60 mg Study Drug', ''],
    'description': ['description 1', ''],
    'label': ['label 1', ''],
    'pharmacologicClass': ['FDA: A=B', ''],
    'administrableDoseForm': ['TABLET', ''],
    'ingredientRole': ['HL7:   100000072072=Active', 'HL7:   100000072072=Active'],
    'substanceName': ['Ingredient C', 'Ingredient D'],
    'substanceDescription': ['description 2', 'description 3'],
    'substanceLabel': ['label 2', 'label 3'],
    'substanceCode': ['', ''],
    'strengthName': ['60 mg', '120 mg'],
    'strengthDescription': ['', ''],
    'strengthLabel': ['', ''],
    'strengthNumerator': ['60 mg', '120 mg'],
    'strengthDenominator': ['1 TABLET', '1 TABLET'],
    'referenceSubstanceName': ['', ''],
    'referenceSubstanceDescription': ['', ''],
    'referenceSubstanceLabel': ['', ''],
    'referenceSubstanceCode': ['', ''],
    'referenceSubstanceStrengthName': ['', ''],
    'referenceSubstanceStrengthDescription': ['', ''],
    'referenceSubstanceStrengthLabel': ['', ''],
    'referenceSubstanceStrengthNumerator': ['', ''],
    'referenceSubstanceStrengthDenominator': ['', '']
  }  
  sheet = _setup_sheet(mocker, globals, data, ids)
  assert sheet.items[0].ingredients[0].substance.name == 'Ingredient C'
  assert sheet.items[0].ingredients[1].substance.name == 'Ingredient D'

def test_create_3(mocker, globals):
  ids = [
     'Id_1',  'Id_2',  'Id_3',  'Id_4',  'Id_5',  'Id_6',  'Id_7',  'Id_8',  'Id_9', 'Id_10', 
    'Id_11', 'Id_12', 'Id_13', 'Id_14', 'Id_15', 'Id_16', 'Id_17', 'Id_18', 'Id_19', 'Id_20',
    'Id_21', 'Id_22', 'Id_23', 'Id_24', 'Id_25', 'Id_26', 'Id_27', 'Id_28', 'Id_29', 'Id_30',
    'Id_31', 'Id_32', 'Id_33', 'Id_34', 'Id_35', 'Id_36', 'Id_37', 'Id_38', 'Id_39', 'Id_40',
  ]
  data = {
    'name': ['60 mg Study Drug', ''],
    'description': ['description 1', ''],
    'label': ['label 1', ''],
    'pharmacologicClass': ['FDA: A=B', ''],
    'administrableDoseForm': ['TABLET', ''],
    'ingredientRole': ['HL7:   100000072072=Active', 'HL7:   100000072072=Active'],
    'substanceName': ['Ingredient C', 'Ingredient D'],
    'substanceDescription': ['description 2', 'description 3'],
    'substanceLabel': ['label 2', 'label 3'],
    'substanceCode': ['', ''],
    'strengthName': ['60 mg', '120 mg'],
    'strengthDescription': ['', ''],
    'strengthLabel': ['', ''],
    'strengthNumerator': ['60 mg', '120 mg'],
    'strengthDenominator': ['1 TABLET', '1 TABLET'],
    'referenceSubstanceName': ['', 'albuterol base'],
    'referenceSubstanceDescription': ['', 'Reference description'],
    'referenceSubstanceLabel': ['', 'Reference label'],
    'referenceSubstanceCode': ['', ''],
    'referenceSubstanceStrengthName': ['', '90 μg'],
    'referenceSubstanceStrengthDescription': ['', 'Ref strength description'],
    'referenceSubstanceStrengthLabel': ['', 'Ref strength label'],
    'referenceSubstanceStrengthNumerator': ['', '90	ug'],
    'referenceSubstanceStrengthDenominator': ['', '1	INHALATION']
  }  
  sheet = _setup_sheet(mocker, globals, data, ids)
  assert sheet.items[0].ingredients[0].substance.referenceSubstance == None
  assert sheet.items[0].ingredients[1].substance.referenceSubstance.name == 'albuterol base'
  assert sheet.items[0].ingredients[1].substance.referenceSubstance.strengths[0].name == '90 μg'

def test_missign_column_error(mocker, globals):
  ids = ['Id_1', 'Id_2', 'Id_3', 'Id_4', 'Id_5', 'Id_6', 'Id_7', 'Id_8', 'Id_9', 'Id_10', 'Id_11', 'Id_12', 'Id_13', 'Id_14']
  data = {
    'name': ['60 mg Study Drug'],
    'description': ['description 1'],
    'label': ['label 1'],
    'administrableDoseForm': ['TABLET'],
    'ingredientRole': ['HL7:   100000072072=Active'],
    'substanceName': ['Ingredient C'],
    'substanceDescription': ['description 2'],
    'substanceLabel': ['label 2'],
    'substanceCode': [''],
    'strengthName': ['60 mg'],
    'strengthDescription': [''],
    'strengthLabel': [''],
    'strengthNumerator': ['60 mg'],
    'strengthDenominator': ['1 TABLET'],
    'referenceSubstanceName': [''],
    'referenceSubstanceDescription': [''],
    'referenceSubstanceLabel': [''],
    'referenceSubstanceCode': [''],
    'referenceSubstanceStrengthName': [''],
    'referenceSubstanceStrengthDescription': [''],
    'referenceSubstanceStrengthLabel': [''],
    'referenceSubstanceStrengthNumerator': [''],
    'referenceSubstanceStrengthDenominator': ['']
  }  
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  sheet = _setup_sheet(mocker, globals, data, ids)
  assert mock_error.call_count == 1
  mock_error.assert_has_calls([mocker.call('studyDesignProducts', None, None, "Exception. Error [Failed to detect column(s) 'pharmacologicClass' in sheet] while reading sheet 'studyDesignProducts'. See log for additional details.", 40)])

def _setup_sheet(mocker, globals, data, ids):
  globals.cross_references.clear()
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=ids
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame.from_dict(data)
  return StudyDesignProductSheet("", globals)



import json
import os
from usdm_excel.globals import Globals
from usdm_excel.code_base import CodeBase

class ISO3166(CodeBase):

  def __init__(self, globals: Globals):
    super().__init__(globals)
    f = open(os.path.join(os.path.dirname(__file__), 'data', 'iso_3166.json'))
    self.db = json.load(f)

  def code(self, code):
    code, decode = self._get_decode(code)
    return self._build(code=code, system='ISO 3166 1 alpha3', version='2020-08', decode=decode)

  def region_code(self, decode):
    code, decode = self._get_region_decode(decode)
    return self._build(code=code, system='ISO 3166 1 alpha3', version='2020-08', decode=decode)

  def _get_decode(self, code):
    if len(code) == 2:
      field = 'alpha-2'
    else:
      field = 'alpha-3'
    entry = next((item for item in self.db if item[field] == code), None)
    if entry == None:
      return 'DNK', 'Denmark'
    else:
      return entry['alpha-3'], entry['name']
    
  def _get_region_decode(self, decode):
    for scope in ['region', 'sub-region', 'intermediate-region']:
      entry = next((item for item in self.db if item[scope].upper() == decode.upper()), None)
      if entry:
        return entry[f'{scope}-code'], entry[scope]
    return '150', 'Europe'
      


from typing import List, Literal
from .syntax_template import SyntaxTemplate

class Condition(SyntaxTemplate):
  contextIds: List[str] = []
  appliesToIds: List[str] = []
  instanceType: Literal['Condition']



import pandas as pd

from usdm_excel.configuration_sheet import ConfigurationSheet
from usdm_excel.option_manager import Options, EmptyNoneOption


def test_defaults(mocker, globals):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {'col_1': ['Option X', 'Option 2', 'Option 3'], 'col_2': ['maybe', 'True', '']}
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data)
  configuration = ConfigurationSheet("", globals)
  assert globals.option_manager.get(Options.EMPTY_NONE) == EmptyNoneOption.NONE.value
  assert globals.template_manager.get('sponsor') == 'document'

def test_template(mocker, globals):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {'col_1': ['template'], 'col_2': ['xxx =    yyy']}
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data)
  configuration = ConfigurationSheet("", globals)
  assert globals.template_manager.get('Xxx') == 'yyy'

def test_use_template_deprecated(mocker, globals):
  mock_sheets = mocker.patch("usdm_excel.base_sheet.BaseSheet._get_sheet_names")
  mock_sheets.side_effect=['document']
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {'col_1': ['USE template'], 'col_2': ['   zzzz   ']}
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data)
  configuration = ConfigurationSheet("", globals)
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "configuration"
  assert mock_error.call_args[0][1] == None
  assert mock_error.call_args[0][2] == None
  assert mock_error.call_args[0][3] == "The USE TEMPLATE option is now deprecated and will be ignored."

def test_usdm_version_deprecated(mocker, globals):
  mock_sheets = mocker.patch("usdm_excel.base_sheet.BaseSheet._get_sheet_names")
  mock_sheets.side_effect=['document']
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {'col_1': ['Usdm VERsion'], 'col_2': ['XXX']}
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data)
  configuration = ConfigurationSheet("", globals)
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "configuration"
  assert mock_error.call_args[0][1] == None
  assert mock_error.call_args[0][2] == None
  assert mock_error.call_args[0][3] == "The USDM VERSION option is now deprecated and will be ignored."

def test_sdr_deprecated(mocker, globals):
  mock_sheets = mocker.patch("usdm_excel.base_sheet.BaseSheet._get_sheet_names")
  mock_sheets.side_effect=['document']
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {'col_1': ['sdr DESCRIPTION'], 'col_2': ['']}
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data)
  configuration = ConfigurationSheet("", globals)
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "configuration"
  assert mock_error.call_args[0][1] == None
  assert mock_error.call_args[0][2] == None
  assert mock_error.call_args[0][3] == "The SDR DESCRIPTION option is now deprecated and will be ignored."

def test_sdr_root_deprecated(mocker, globals):
  mock_sheets = mocker.patch("usdm_excel.base_sheet.BaseSheet._get_sheet_names")
  mock_sheets.side_effect=['document']
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {'col_1': ['SDR root'], 'col_2': ['']}
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data)
  configuration = ConfigurationSheet("", globals)
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "configuration"
  assert mock_error.call_args[0][1] == None
  assert mock_error.call_args[0][2] == None
  assert mock_error.call_args[0][3] == "The SDR ROOT option is now deprecated and will be ignored."

def test_set_prev_next_deprecated(mocker, globals):
  mock_sheets = mocker.patch("usdm_excel.base_sheet.BaseSheet._get_sheet_names")
  mock_sheets.side_effect=['document']
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {'col_1': ['SDR prev next'], 'col_2': ['']}
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data)
  configuration = ConfigurationSheet("", globals)
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "configuration"
  assert mock_error.call_args[0][1] == None
  assert mock_error.call_args[0][2] == None
  assert mock_error.call_args[0][3] == "The SDR PREV NEXT option is now deprecated and will be ignored."



from yattag import Doc
from usdm_excel.document.template_base import TemplateBase
from usdm_excel.document.soa import SoA
from usdm_excel.base_sheet import BaseSheet
from usdm_model.schedule_timeline import ScheduleTimeline
from usdm_model.study import Study

class TemplatePlain(TemplateBase):

  # def __init__(self, parent: BaseSheet, study: Study):
  #   super().__init__(parent, study)

  def title_page(self, attributes):
    doc = Doc()
    with doc.tag('table'):
      self._title_page_entry(doc, 'Full Title:', f'{self._elements.study_full_title()}')
      self._title_page_entry(doc, 'Trial Acronym:', f'{self._elements.study_acronym()}')
      self._title_page_entry(doc, 'Protocol Identifier:', f'{self._elements.study_identifier()}')
      self._title_page_entry(doc, 'Version Number:', f'{self._elements.study_version_identifier()}')
      self._title_page_entry(doc, 'Version Date:', f'{self._elements.study_approval_date()}')
      self._title_page_entry(doc, 'Amendment Identifier:', f'{self._elements.amendment()}')
      self._title_page_entry(doc, 'Amendment Scope:', f'{self._elements.amendment_scopes()}')
      self._title_page_entry(doc, 'Trial Phase:', f'{self._elements.study_phase()}')
      self._title_page_entry(doc, 'Short Title:', f'{self._elements.study_short_title()}')
      self._title_page_entry(doc, 'Sponsor Name and Address:', f'{self._elements.organization_name_and_address()}')
      self._title_page_entry(doc, 'Regulatory Agency Identifier Number(s):', f'{self._elements.study_regulatory_identifiers()}')
      self._title_page_entry(doc, 'Spondor Approval Date:', f'{self._elements.approval_date()}')
    result = doc.getvalue()
    return result

  def inclusion(self, attributes: dict):  
    return self._criteria("C25532")
  
  def exclusion(self, attributes: dict):  
    return self._criteria("C25370")

  def objective_endpoints(self, attributes: dict):
    doc = Doc()
    with doc.tag('table', klass='table'):
      for item in self._objective_endpoints_list():
        self._objective_endpoints_entry(doc, item['objective'], item['endpoints'])
    return doc.getvalue()

  def soa(self, attributes: dict):
    try:
      doc = Doc()
      for timeline in self._study_design.scheduleTimelines:      
        doc.asis(self.timeline({'timeline': timeline}))
    except Exception as e:
      self.parent._general_exception(f"Error raised generating SoA", e)
    return doc.getvalue()

  def timeline(self, attributes: dict):
    try:
      doc = Doc()
      timeline = None
      timeline = self._resolve_timeline(attributes)
      footnote = 1
      footnotes = []
      soa = SoA(self.parent, self._study_design, timeline)
      result = soa.generate()
      with doc.tag('div', klass="page soa-page table-responsive"):
        with doc.tag('p'):
          with doc.tag('b'):
            doc.asis(f"Timeline: {timeline.label}, {timeline.entryCondition}")
        with doc.tag('table', klass='table table-bordered table-sm', style="width:100%"):
          for row in range(len(result)):
            with doc.tag('tr'):
              for col in range(len(result[row])):
                if 'set' in result[row][col].keys():
                  label = 'X' if result[row][col]['set'] else ''
                else:
                  label = result[row][col]['label']
                with doc.tag('td'):
                  klass = 'soa-activity-text' if col == 0 else 'soa-body-text'
                  with doc.tag('p', klass=klass):
                    doc.asis(label)
                    if 'condition' in result[row][col].keys():
                      with doc.tag('sup'):
                        footnote_str = str(footnote)
                        doc.text(footnote_str)
                      footnotes.append({'number': footnote_str, 'text': result[row][col]['condition'].text})
                      footnote += 1
        if footnotes:
          with doc.tag('table', klass='table table-borderless table-sm'):
            for item in footnotes:
              with doc.tag('tr'):
                with doc.tag('td'):
                  with doc.tag('p', klass="soa-footnote-text"):
                    with doc.tag('sup'):
                      doc.text(item['number'])
                with doc.tag('td'):
                  with doc.tag('p', klass="soa-footnote-text"):
                    doc.asis(item['text'])
    except Exception as e:
      if timeline:
        self.parent._general_exception(f"Error raised generating timeline '{timeline.name}'", e)
        return "Error encountered creating timeline {timeline.name}"  
      else:  
        self.parent._general_exception(f"Error raised generating timeline, name not available", e)
        return "Error encountered creating timeline"  
    return doc.getvalue()

  def _resolve_timeline(self, attributes):
    return attributes['timeline'] if isinstance(attributes['timeline'], ScheduleTimeline) else self.parent.globals.cross_references.get(ScheduleTimeline, attributes['timeline'])


  def _criteria(self, type):
    doc = Doc()
    with doc.tag('table', klass='table'):
      for criterion in self._criteria_list(type):
        self._critieria_entry(doc, criterion['identifier'], criterion['text'])
    return doc.getvalue()

  def _critieria_entry(self, doc, identifier, entry):
    with doc.tag('tr'):
      with doc.tag('td'):
        self._add_checking_for_tag(doc, 'p', identifier)
      with doc.tag('td'):
        self._add_checking_for_tag(doc, 'p', entry)

  def _objective_endpoints_entry(self, doc, objective, endpoints):
    with doc.tag('tr'):
      with doc.tag('td'):
        self._add_checking_for_tag(doc, 'p', objective)
      with doc.tag('td'):
        for endpoint in endpoints:
          self._add_checking_for_tag(doc, 'p', endpoint)

  def _title_page_entry(self, doc, title, entry):
    with doc.tag('tr'):
      with doc.tag('th'):
        self._add_checking_for_tag(doc, 'p', title)
      with doc.tag('td'):
        self._add_checking_for_tag(doc, 'p', entry)

  def _criteria_list(self, type):
    results = []
    criteria = {x.id:x for x in self._study_version.criteria}
    items = [criteria[c] for c in self._study_design.population.criterionIds if criteria[c].category.code == type ]
    items.sort(key=lambda d: d.identifier)
    for item in items:
      result = {'identifier': item.identifier, 'text': self._reference(item, 'text')}
      results.append(result)
    return results

  def _objective_endpoints_list(self):
    results = []
    for item in self._study_design.objectives:
      result = {'objective': self._reference(item, 'text'), 'endpoints': []}
      for endpoint in item.endpoints:
        result['endpoints'].append(self._reference(endpoint, 'text'))
      results.append(result)
    return results



from typing import List, Literal
from .api_base_model import ApiBaseModelWithId
from .code import Code

class AliasCode(ApiBaseModelWithId):
  standardCode: Code
  standardCodeAliases: List[Code] = []
  instanceType: Literal['AliasCode']



from usdm_excel.template_manager import *

def test_create(globals):
  object = TemplateManager(globals)
  assert len(object._items.keys()) == 1
  assert object._items == {'SPONSOR': 'document'}

def test_default(globals):
  object = TemplateManager(globals)
  assert object.default_template == 'SPONSOR'

def test_add(globals):
  globals.template_manager._items = {}
  globals.template_manager.add('fred', 'value')
  assert len(globals.template_manager._items.keys()) == 1
  assert globals.template_manager._items['FRED'] == 'value'
  globals.template_manager.add('Sid1', 'value')
  assert len(globals.template_manager._items.keys()) == 2
  assert globals.template_manager._items['SID1'] == 'value'

def test_get(globals):
  globals.template_manager._items = {}
  globals.template_manager._items['FRED'] = 'value'
  assert globals.template_manager.get('fred') == 'value'
  assert globals.template_manager.get('FRED') == 'value'

def test_includes_and_templates(globals):
  globals.template_manager._items = {}
  globals.template_manager._items['FRED'] = 'value_f'
  globals.template_manager._items['XXX'] = 'value_x'
  assert globals.template_manager.includes('fred') == True
  assert globals.template_manager.includes('FRED') == True
  assert globals.template_manager.includes('FRED1') == False
  assert globals.template_manager.includes('XXX') == True
  assert globals.template_manager.includes('XX X') == False
  assert globals.template_manager.templates() == ['FRED', 'XXX']

def test_clear(globals):
  globals.template_manager._items = {}
  globals.template_manager._items['fred'] = 'value'
  assert len(globals.template_manager._items.keys()) == 1
  globals.template_manager.clear()
  assert len(globals.template_manager._items.keys()) == 0

def test_tidy_remove_default_1(globals):
  object = TemplateManager(globals)
  object.add('fred', 'value_fred')
  object.add('Sid', 'value_sid')
  object.tidy(['value_fred', 'value_sid'])
  assert len(object._items.keys()) == 2
  assert object._items['FRED'] == 'value_fred'
  assert object._items['SID'] == 'value_sid'

def test_tidy_remove_default_2(globals):
  object = TemplateManager(globals)
  object.add('fred', 'value_fred')
  object.add('Sid', 'value_sid')
  object.tidy(['value_fred', 'value_sid', 'document'])
  assert len(object._items.keys()) == 3
  assert object._items['FRED'] == 'value_fred'
  assert object._items['SID'] == 'value_sid'
  assert object._items['SPONSOR'] == 'document'

def test_utility(globals):
  keys = ['SPONSOR', 'FRED', 'SID']
  values = ['document', 'value_fred', 'value_sid']
  object = TemplateManager(globals)
  object.add('fred', 'value_fred')
  object.add('Sid', 'value_sid')
  assert list(object.keys()) == keys
  assert list(object.values()) == values
  index = 0
  for key, value in object.items():
    assert key == keys[index]
    assert value == values[index]
    index += 1
  assert [x for x in list(object.values())] == values


from typing import List, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .code import Code
from .comment_annotation import CommentAnnotation

class StudyArm(ApiBaseModelWithIdNameLabelAndDesc):
  type: Code
  dataOriginDescription: str
  dataOriginType: Code
  populationIds: List[str] = []
  notes: List[CommentAnnotation] = []
  instanceType: Literal['StudyArm']



import traceback
from usdm_excel.base_sheet import BaseSheet
from usdm_model.condition import Condition
from usdm_excel.globals import Globals

class StudyDesignConditionSheet(BaseSheet):

  SHEET_NAME = 'studyDesignConditions'

  def __init__(self, file_path: str, globals: Globals):
    try:
      self.items = []
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME, optional=True)
      if self.success:
        for index, row in self.sheet.iterrows():
          name = self.read_cell_by_name(index, 'name')
          description = self.read_cell_by_name(index, 'description')
          label = self.read_cell_by_name(index, 'label')
          text = self.read_cell_by_name(index, 'text')
          context = self.read_cell_by_name(index, 'context')
          context_refs = self._process_context_references(context, index)
          applies_to = self.read_cell_by_name(index, 'appliesTo')
          applies_refs = self._process_applies_to_references(applies_to, index)
          params = {'name': name, 'description': description, 'label': label, 'text': text, 'appliesToIds': applies_refs, 'contextIds': context_refs}
          item = self.create_object(Condition, params)
          if item:
            self.items.append(item)
            self.globals.cross_references.add(name, item)     
    except Exception as e:
      self._sheet_exception(e)

  def _process_context_references(self, references_list, index):
    return self._process_references(references_list, ['ScheduledActivityInstance', 'Activity'], index, 'context', False)
  
  def _process_applies_to_references(self, references_list, index):
    return self._process_references(references_list, ['Procedure', 'Activity', 'BiomedicalConcept', 'BiomedicalConceotCategory', 'BiomedicalConceptSurrogate'], index, 'appliesTo')
  
  def _process_references(self, references_list, klasses, index, column_name, references_required=True):
    references = [x.strip() for x in self._state_split(references_list)]
    #references = [x.strip() for x in references_list.split(',')]
    results = []
    for reference in references:
      if reference:
        found = False
        for klass in klasses:
          xref = self.globals.cross_references.get(klass, reference)
          if xref:
            results.append(xref.id)
            found = True
            break
        if not found:
          self._error(index, self._get_column_index(column_name), f"Could not resolve condition reference '{reference}'")
    if not results and references_required:
      self._error(index, self._get_column_index(column_name), f"No condition references found for '{references_list}', at least one required")
    return results



import warnings
import traceback
from bs4 import BeautifulSoup   

def usdm_reference(item, attribute):
  return f'<usdm:ref klass="{item.__class__.__name__}" id="{item.id}" attribute="{attribute}"></usdm:ref>'

def get_soup(text, parent):
  try:
    with warnings.catch_warnings(record=True) as warning_list:
      result =  BeautifulSoup(text, 'html.parser')
    if warning_list:
      for item in warning_list:
        parent._general_debug(f"Warning raised within Soup package, processing '{text}'\nMessage returned '{item.message}'")
    return result
  except Exception as e:
    parent._exception(f"Eerror raised parsing '{text}'", e)
    return ""
  


from typing import Literal, Union
from .syntax_template import SyntaxTemplate
from .code import Code

class EligibilityCriterion(SyntaxTemplate):
  category: Code
  identifier: str
  nextId: Union[str, None] = None
  previousId: Union[str, None] = None
  instanceType: Literal['EligibilityCriterion']



import traceback
from usdm_excel.base_sheet import BaseSheet
from usdm_model.study_epoch import StudyEpoch
from usdm_model.study_arm import StudyArm
from usdm_model.study_element import StudyElement
from usdm_model.study_cell import StudyCell
from usdm_model.study_design import StudyDesign
from usdm_model.masking import Masking
from usdm_excel.alias import Alias
from usdm_excel.option_manager import *
from usdm_excel.cdisc_ct import CDISCCT
from usdm_excel.globals import Globals

class StudyDesignSheet(BaseSheet):

  SHEET_NAME = 'studyDesign'
  
  NAME_LABEL = ['studyDesignName', 'name']
  DESCRIPTION_LABEL = ['studyDesignDescription', 'description']
  LABEL_LABEL = ['label']
  TA_LABEL = ['therapeuticAreas']
  RATIONALE_LABEL = ['studyDesignRationale']
  BLINDING_LABEL = ['studyDesignBlindingScheme']
  INTENT_LABEL = ['trialIntentTypes']
  TYPES_LABEL = ['trialTypes']
  INT_LABEL = ['interventionModel']
  MAIN_TIMELINE_LABEL = ['mainTimeline']
  OTHER_TIMELINES_LABEL = ['otherTimelines']
  MASKING_ROLE_LABEL = ['masking']
  CHARACTERISTICS_LABEL = ['characteristics']

  PARAMS_NAME_COL = 0
  PARAMS_DATA_COL = 1

  def __init__(self, file_path: str, globals: Globals):
    try:
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME, header=None)
      self.name = "TEST"
      self.description = "An Microsoft Excel test study design"
      self.label=""
      self.epochs = []
      self.epoch_names = {}
      self.arms = []
      self.arm_names = {}
      self.cells = []
      self.elements = {}
      self.study_designs = []
      self.therapeutic_areas = []
      self.rationale = ""
      self.blinding = None
      self.trial_intents = []
      self.characteristics = []
      self.trial_types = []
      self.intervention_model = None
      self.main_timeline = None
      self.other_timelines = []
      self.masks = []
      self.process_sheet()
    except Exception as e:
      self._sheet_exception(e)

  def process_sheet(self):
    general_params = True
    resolved_epochs = [None] * self.sheet.shape[1] 
    resolved_arms = [None] * self.sheet.shape[0] 
    for rindex, row in self.sheet.iterrows():
      key = self.read_cell(rindex, self.PARAMS_NAME_COL)
      if general_params:
        #print(f"KEY: {key}")
        if key in self.NAME_LABEL:
          self.name = self.read_cell(rindex, self.PARAMS_DATA_COL)
        elif key in self.LABEL_LABEL:
          self.label = self.read_cell(rindex, self.PARAMS_DATA_COL)
        elif key in self.DESCRIPTION_LABEL:
          self.description = self.read_cell(rindex, self.PARAMS_DATA_COL)
        elif key in self.TA_LABEL:
          self.therapeutic_areas = self.read_other_code_cell_mutiple(rindex, self.PARAMS_DATA_COL)
        elif key in self.RATIONALE_LABEL:
          self.rationale = self.read_cell(rindex, self.PARAMS_DATA_COL)
        elif key in self.BLINDING_LABEL:
          self.blinding = Alias(self.globals).code(self.read_cdisc_klass_attribute_cell('StudyDesign', 'studyDesignBlindingScheme',rindex, self.PARAMS_DATA_COL), [])
        elif key in self.INTENT_LABEL:
          self.trial_intents = self.read_cdisc_klass_attribute_cell_multiple('StudyDesign', 'trialIntentType', rindex, self.PARAMS_DATA_COL)
        elif key in self.CHARACTERISTICS_LABEL:
          self.characteristics = self.read_cdisc_klass_attribute_cell_multiple('StudyDesign', 'characteristics', rindex, self.PARAMS_DATA_COL)
          #print(f"CHARAC: {self.characteristics}")
        elif key in self.TYPES_LABEL:
          self.trial_types = self.read_cdisc_klass_attribute_cell_multiple('StudyDesign', 'trialType', rindex, self.PARAMS_DATA_COL)
        elif key in self.INT_LABEL:
          self.intervention_model = self.read_cdisc_klass_attribute_cell('StudyDesign', 'interventionModel', rindex, self.PARAMS_DATA_COL)
        elif key in self.MAIN_TIMELINE_LABEL:
          #print(f"MAIN TL: {rindex}")
          self.main_timeline = self.read_cell(rindex, self.PARAMS_DATA_COL)
        elif key in self.OTHER_TIMELINES_LABEL:
          #print(f"OTHER TL: {rindex}")
          self.other_timelines = self.read_cell_multiple(rindex, self.PARAMS_DATA_COL)
        elif key in self.MASKING_ROLE_LABEL:
          #print(f"MASKING: {rindex}")
          self._set_masking(rindex, self.PARAMS_DATA_COL)
        elif key in '':
          general_params = False
          start_row = rindex + 1
          #print(f"START: {start_row}")
      else:
        for cindex in range(0, len(self.sheet.columns)):
          epoch_index = cindex - 1
          cell = self.read_cell(rindex, cindex)
          #print(f"ARMS EPOCHS: {rindex} = {cell}")
          if rindex == start_row:
            if cindex != 0:
              resolved_epochs[epoch_index] = self._add_epoch(cell)
          else:
            arm_index = rindex - start_row - 1
            if cindex == 0:
              resolved_arms[arm_index] = self._add_arm(cell)
            else:
              cell_elements = []
              element_names = self.read_cell_multiple(rindex, cindex)
              for name in element_names:
                element = self._add_element(name)
                if element is not None:
                  cell_elements.append(element.id)
              cell_arm = resolved_arms[arm_index].id
              cell_epoch = resolved_epochs[epoch_index].id
              if cell_arm is not None and cell_epoch is not None:
                self.cells.append(self._add_cell(arm=cell_arm, epoch=cell_epoch, elements=cell_elements))
              else:
                self._general_error(f"Cannot resolve arm and/or epoch for cell [{arm_index + 1},{epoch_index + 1}]")
                
      self.double_link(self.epochs, 'previousId', 'nextId')

    study_design = self._create_design()
    self.study_designs.append(study_design)

  def _add_arm(self, name):
    arm = self.globals.cross_references.get(StudyArm, name)
    if arm is not None:
      if name not in self.arm_names:
        self.arm_names[name] = True
        self.arms.append(arm)
      return arm
    else:
      self._general_error(f"No arm definition found for arm '{name}'")
      return None

  def _add_epoch(self, name):
    epoch = self.globals.cross_references.get(StudyEpoch, name)
    if epoch is not None:
      if name not in self.epoch_names:
        self.epoch_names[name] = True
        self.epochs.append(epoch)
      return epoch
    else:
      self._general_error(f"No epoch definition found for epoch '{name}'")
      return None
  
  def _add_element(self, name):
    element = self.globals.cross_references.get(StudyElement, name)
    if element is not None:
      if name not in self.elements:
        self.elements[name] = element
      return element
    else:
      self._general_error(f"No element definition found for element '{name}'")
      return None

  def _add_cell(self, arm, epoch, elements):
    try:
      return StudyCell(
        id=self.globals.id_manager.build_id(StudyCell), 
        armId=arm,
        epochId=epoch,
        elementIds=elements
      )
    except Exception as e:
      self._general_error("Failed to create StudyCell object", e)
      return None

  def _create_design(self):
    try:
      result = StudyDesign(
        id=self.globals.id_manager.build_id(StudyDesign), 
        name=self.name,
        description=self.description,
        label=self.label,
        trialIntentTypes=self.trial_intents, 
        trialTypes=self.trial_types, 
        interventionModel=self.intervention_model,
        studyCells=self.cells,
        arms=self.arms,
        epochs=self.epochs,
        elements=list(self.elements.values()),
        therapeuticAreas=self.therapeutic_areas,
        rationale=self.rationale, 
        blindingSchema=self.blinding, 
        contents=[],
        maskingRoles=self.masks,
        characteristics=self.characteristics
      )
      #print(f"SD: {result.characteristics}")
      return result
    except Exception as e:
      self._general_error("Failed to create StudyDesign object", e)
      return None
  

  def _set_masking(self, rindex, cindex):
    # if self.globals.option_manager.get(Options.USDM_VERSION) == '2':
    #   return None
    # else:
    try:
      text = self.read_cell(rindex, cindex)
      parts = text.split('=')
      if len(parts) == 2: 
        code = CDISCCT(self.globals).code_for_attribute('Masking', 'role', parts[0].strip())
        if code:
          mask = Masking(id=self.globals.id_manager.build_id(Masking), description=parts[1].strip(), role=code)
          self.masks.append(mask)
          self.globals.cross_references.add(mask.id, mask)
          return mask
        else:
          self._error(rindex, cindex, f"Failed to decode masking role data '{text}', must be a valid role code '{parts[0]}'")
          return None
      else:
        self._error(rindex, cindex, f"Failed to decode masking role data '{text}', no '=' detected")
        return None
    except Exception as e:
      self._error(rindex, cindex, "Failed to create Masking object", e)
      return None



import pytest
import pandas as pd
from usdm_excel.study_design_population_sheet.study_design_population_sheet import StudyDesignPopulationSheet
from usdm_model.characteristic import Characteristic

xfail = pytest.mark.xfail
def test_create(mocker, globals):
  mock_cross_ref = mocker.patch("usdm_excel.cross_ref.CrossRef.get")
  mock_cross_ref.side_effect=[
    Characteristic(id="CH_1", name="CHAR_1", text='Something', instanceType='Characteristic'), 
    Characteristic(id="CH_2", name="CHAR_2", text='Something', instanceType='Characteristic'), 
    Characteristic(id="CH_3", name="CHAR_3", text='Something', instanceType='Characteristic')
  ]
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['R_1', 'R_2', 'X_3', 'R_4', 'Code_5', 'PopulationId_1', 'X_7', 'X_8', 'X_9', 'X_10', 'X_11', 'CohortId_1', 'X_13', 'X_14', 'X_15', 'X_16', 'X_17', 'CohortId_2', 'X_18', 'X_19', 'X_20']
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['MAIN',   'POP01', 'Main pop', 'Main Pop', 'BOTH',   '10..20', '100..110', '100..110 years', 'Y', ""], 
    ['COHORT', 'POP02', 'Cohort 1', 'Cohort 1', 'MALE',   '5..10',  '50..50',   '50..50 years',   'Y', "CHAR_1, CHAR_2"], 
    ['COHORT', 'POP03', 'Cohort 2', 'Cohort 2', 'FEMALE', '5..10',  '50..60',   '50..60 years',   'Y', "CHAR_3"], 
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['level', 'name', 'description', 'label', 'plannedSexOfParticipants', "plannedCompletionNumber", "plannedEnrollmentNumber", "plannedAge", 'includesHealthySubjects', 'characteristics'])
  item = StudyDesignPopulationSheet("", globals)
  assert item.population.id == 'PopulationId_1'
  assert item.population.name == 'POP01'
  assert item.population.description == 'Main pop'
  assert item.population.label == 'Main Pop'
  assert item.population.plannedEnrollmentNumber.id == 'R_2'
  assert item.population.plannedCompletionNumber.id == 'R_1'
  assert item.population.plannedSex[0].id == 'Code_5'
  assert item.population.plannedSex[0].decode == 'Both'
  assert item.population.plannedAge.id == 'R_4'
  assert item.population.cohorts[0].id == 'CohortId_1'
  assert item.population.cohorts[0].name == 'POP02'
  assert item.population.cohorts[0].characteristics[0].id == 'CH_1'
  assert item.population.cohorts[0].characteristics[1].id == 'CH_2'
  assert item.population.cohorts[1].id == 'CohortId_2'
  assert item.population.cohorts[1].name == 'POP03'
  assert item.population.cohorts[1].characteristics[0].id == 'CH_3'
  
def test_create_empty(mocker, globals):
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['category', 'identifier', 'name', 'description', 'label', 'text', 'dictionary'])
  items = StudyDesignPopulationSheet("", globals)
  assert items.population == None

def test_read_cell_by_name_error(mocker, globals):
  globals.cross_references.clear()
  call_parameters = []
  
  def my_add(sheet, row, column, message, level=10):
    call_parameters.append((sheet, row, column, message, level))
    return None

  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add", side_effect=my_add)
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['MAIN',   'POP01', 'Main pop', 'Main Pop', '10..20', '100..110', '100..110 years', 'Y', ""]
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['level', 'name', 'description', 'label', "plannedCompletionNumber", "plannedEnrollmentNumber", "plannedAge", 'includesHealthySubjects', 'characteristics'])
  items = StudyDesignPopulationSheet("", globals)
  mock_error.assert_called()
  assert call_parameters == [
    ('studyDesignPopulations', None, None, "Exception. Error [Failed to detect column(s) 'plannedSexOfParticipants' in sheet] while reading sheet 'studyDesignPopulations'. See log for additional details.", 40)
  ]
  



import json
import traceback
from uuid import uuid4
from usdm_excel import USDMExcel
from usdm_model.wrapper import Wrapper
from usdm_db.document.document import Document
from usdm_db.fhir.to_fhir import ToFHIR
from usdm_db.fhir.from_fhir import FromFHIR
from usdm_db.errors_and_logging.errors_and_logging import ErrorsAndLogging
from usdm_db.neo4j_dict import Neo4jDict
from usdm_db.timeline import Timeline

class USDMDb():

  SYSTEM_NAME = "CDISC E2J"

  FULL_HTML = Timeline.FULL
  BODY_HTML = Timeline.BODY
  
  def __init__(self):
    self._wrapper = None
    self._excel = None
    self._errors_and_logging = ErrorsAndLogging()

  def errors(self):
    return self._errors_and_logging.errors().dump()
  
  def wrapper(self):
    return self._wrapper
  
  def excel(self):
    return self._excel

  def from_json(self, data):
    self._wrapper = Wrapper.model_validate(data)
    
  def from_excel(self, file_path):
    self._excel = USDMExcel(file_path)
    self._wrapper = self._excel.execute()
    return self._excel.errors()

  def default_template(self):
    return self._excel.default_template()
  
  def templates(self):
    return self._excel.templates()
  
  def from_fhir(self, data: str):
    fhir = FromFHIR(self._errors_and_logging)
    self._wrapper = fhir.from_fhir(data)
    return True
  
  def was_m11(self) -> bool:
    self._errors_and_logging.deprecated("Method 'was_m11' deprecated, use 'is_m11' going forward")
    return self._excel.was_m11()

  def is_m11(self) -> bool:
    return self._excel.is_m11()

  def to_json(self):
    try:
      raw_json = self._wrapper.to_json()
    except Exception as e:
      message = self._format_exception("Failed to generate JSON output", e)
      raw_json = json.dumps({'error': message}, indent = 2)
    return raw_json

  def to_fhir(self, template):
    try:
      study = self._wrapper.study
      fhir = ToFHIR(study, template)
      raw_json = fhir.to_fhir(uuid4())
    except Exception as e:
      message = self._format_exception("Failed to generate FHIR output", e)
      raw_json = json.dumps({'error': message}, indent = 2)
    return raw_json

  def to_html(self, template_name: str, highlight: bool=False):
    try:
      study = self._wrapper.study
      doc = Document(study, template_name, self._errors_and_logging)
      html = doc.to_html(highlight)
    except Exception as e:
      message = self._format_exception("Failed to generate HTML output", e)
      html = f"<p>{message}</p>"
    return html

  def to_pdf(self, template_name: str, test: bool=True):
    try:
      study = self._wrapper.study
      doc = Document(study, template_name, self._errors_and_logging)
      bytes = doc.to_pdf(test)
    except Exception as e:
      message = self._format_exception("Failed to generate PDF output", e)
      bytes = bytearray()
      bytes.extend(map(ord, message))    
    return bytes

  def to_neo4j_dict(self):
    return Neo4jDict(self._wrapper.study).to_dict()

  def to_timeline(self, level=FULL_HTML):
    return Timeline(self._wrapper.study).to_html(level)

  def _format_exception(self, message, e):
    return f"{message}, exception {e}\n{traceback.format_exc()}"



from usdm_excel.id_manager import IdManager
from usdm_excel.cross_ref import CrossRef
from usdm_excel.ct_version_manager import CTVersionManager
from usdm_excel.template_manager import TemplateManager
from usdm_excel.errors_and_logging.errors_and_logging import ErrorsAndLogging
from usdm_excel.option_manager import OptionManager
from usdm_excel.cdisc_ct_library import CDISCCTLibrary
from usdm_excel.cdisc_bc_library import CDISCBCLibrary

class Globals():

  def __init__(self):
    self.errors_and_logging = None
    self.id_manager = None
    self.ct_version_manager = None
    self.template_manager = None
    self.option_manager = None
    self.cross_references = None
    self.cdisc_ct_library = None
    self.cdisc_bc_library = None

  def create(self):
    self.errors_and_logging = ErrorsAndLogging()
    self.id_manager = IdManager(self.errors_and_logging)
    self.ct_version_manager = CTVersionManager(self.errors_and_logging)
    self.template_manager = TemplateManager(self.errors_and_logging)
    self.option_manager = OptionManager(self.errors_and_logging)
    self.cross_references = CrossRef(self.errors_and_logging)
    self.cdisc_ct_library = CDISCCTLibrary(self.errors_and_logging)
    self.cdisc_bc_library = CDISCBCLibrary(self.errors_and_logging, self.cdisc_ct_library, self.id_manager)




import pytest
import pandas as pd
from usdm_excel.annotation_sheet.annotation_sheet import AnnotationSheet
from usdm_model.code import Code

xfail = pytest.mark.xfail

def test_create(mocker, globals):
  data = {
    'name': ['A1', 'A2', 'A3'], 
    'text': ['Annotation Text One', 'Annotation Text Two', 'Annotaiton Text Three'], 
    'codes': ['Sponsor:X=Y', 'SPONSOR: AAA=BBB', 'SPONSOR: WWW=1234, SPONSOR: EEE=3456']
  }
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['Code_1', 'CommentAnnotation_1', 'Code_2', 'CommentAnnotation_2', 'Code_3', 'Code_4', 'CommentAnnotation_3']
  _setup(mocker, globals, data)
  expected_1 = Code(id='Code_1', code='X', codeSystem='SPONSOR', codeSystemVersion='', decode="Y")
  expected_2 = Code(id='Code_2', code='AAA', codeSystem='SPONSOR', codeSystemVersion='', decode="BBB")
  expected_3 = Code(id='Code_3', code='WWW', codeSystem='SPONSOR', codeSystemVersion='', decode="1234")
  expected_4 = Code(id='Code_4', code='EEE', codeSystem='SPONSOR', codeSystemVersion='', decode="3456")
  mock_code = mocker.patch("usdm_excel.cdisc_ct.CDISCCT.code_for_attribute")
  mock_code.side_effect=[expected_1, expected_2, expected_3, expected_4]
  item = AnnotationSheet("", globals)
  assert len(item.items) == 3
  assert item.items[0].id == 'CommentAnnotation_1'
  assert item.items[1].id == 'CommentAnnotation_2'
  assert item.items[1].text == 'Annotation Text Two'
  assert item.items[1].codes == [expected_2]
  assert item.items[2].id == 'CommentAnnotation_3'
  assert item.items[2].codes == [expected_3, expected_4]
  
def test_create_empty(mocker, globals):
  data = {}
  _setup(mocker, globals, data)
  item = AnnotationSheet("", globals)
  assert len(item.items) == 0

def test_read_cell_by_name_error(mocker, globals):
  data = {
    'name': ['A1'], 
    'codes': ['Sponsor:X=Y']
  }
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['Code_1', 'CommentAnnotation_1']
  _setup(mocker, globals, data)
  item = AnnotationSheet("", globals)
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "notes"
  assert mock_error.call_args[0][1] == 1
  assert mock_error.call_args[0][2] == -1
  assert mock_error.call_args[0][3] == "Error attempting to read cell 'text'. Exception: Failed to detect column(s) 'text' in sheet"
  
def _setup(mocker, globals, data):
  globals.cross_references.clear()
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame.from_dict(data)



from usdm_excel.globals import Globals

def test_create():
  globals = Globals()
  globals.create()
  assert globals.errors_and_logging is not None
  assert globals.id_manager is not None
  assert globals.ct_version_manager is not None
  assert globals.option_manager is not None
  assert globals.cross_references is not None
  assert globals.cdisc_ct_library is not None
  assert globals.template_manager is not None



from typing import List, Union, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .comment_annotation import CommentAnnotation

class SyntaxTemplate(ApiBaseModelWithIdNameLabelAndDesc):
  text: str
  dictionaryId: Union[str, None] = None
  notes: List[CommentAnnotation] = []
  instanceType: Literal['SyntaxTemplate']



import re
from usdm_excel.globals import Globals
from usdm_excel.cdisc_ct import CDISCCT

class RangeType():

  def __init__(self, range_info: str, globals: Globals, units_reqd: bool=True, allow_empty: bool=True):
    try:
      self.upper = None
      self.lower = None
      self.units = None
      self.units_code = None
      self.errors = []
      self.empty = False
      self.label = range_info.strip()
      if range_info:
        match = re.match(r"(?P<lower>[+|-]*\d+)(\s*\.\.\s*(?P<upper>[+|-]*\d+))?( \s*(?P<units>.+))?", self.label)
        if match :
          parts = match.groupdict()
          self.lower = parts['lower'].strip()
          if parts['upper']:
            self.upper = parts['upper'].strip()
          else:
            self.upper = self.lower
          if units_reqd:
            if parts['units']:
              self.units = parts['units'].strip()
              self.units_code = CDISCCT(globals).code_for_unit(self.units)
              if not self.units_code:        
                self.errors.append(f"Unable to set the units code for the range '{range_info}'")
            else:
              self.errors.append(f"Could not decode the range value, possible typographical errors '{range_info}'")
        else:
          self.errors.append(f"Could not decode the range value '{range_info}'")
      elif not allow_empty:
        self.errors.append(f"Could not decode the range value, appears to be empty '{range_info}'")
      else:
        self.empty = True
    except Exception as e:
      self.errors.append(f"Exception '{e}' raised decoding range '{range_info}")




import logging
import traceback
import warnings
from usdm_db.errors_and_logging.errors import Errors

class ErrorsAndLogging():

  WARNING = Errors.WARNING
  ERROR = Errors.ERROR
  DEBUG = Errors.DEBUG
  INFO = Errors.INFO

  def __init__(self):
    self._logger = logging.getLogger(__name__)
    self._logger.addHandler(logging.NullHandler())
    self._errors = Errors()

  def errors(self):
    return self._errors
  
  def debug(self, message: str):
    self._logger.debug(message)

  def info(self, message: str):
    self._logger.info(message)

  def exception(self, message: str, e: Exception):
    self._errors.add(f"Exception. {message}. See log for additional details.", self._errors.ERROR)
    self._logger.error(f"Exception '{e}' raised\n\n{message}\n\n{traceback.format_exc()}")

  def warning(self, message: str):
    self._errors.add(message, self._errors.WARNING)
    self._logger.warning(message)

  def error(self, message: str):
    self._errors.add(message, self._errors.ERROR)
    self._logger.error(message)

  def deprecated(self, message: str):
    warnings.warn(message, DeprecationWarning)    
    self.warning(message)


from typing import List, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .code import Code
from .comment_annotation import CommentAnnotation

class Indication(ApiBaseModelWithIdNameLabelAndDesc):
  codes: List[Code] = []
  isRareDisease: bool
  notes: List[CommentAnnotation] = []
  instanceType: Literal['Indication']



from typing import List, Literal, Union
from .syntax_template import SyntaxTemplate
from .code import Code
from .endpoint import Endpoint

class Objective(SyntaxTemplate):
  level: Code
  endpoints: List[Endpoint] = []
  instanceType: Literal['Objective']



from usdm_excel.base_sheet import BaseSheet
from usdm_excel.alias import Alias
from usdm_model.administrable_product import AdministrableProduct
from usdm_model.ingredient import Ingredient
from usdm_model.substance import Substance
from usdm_model.strength import Strength
from usdm_excel.globals import Globals

class StudyDesignProductSheet(BaseSheet):

  SHEET_NAME = 'studyDesignProducts'

  def __init__(self, file_path: str, globals: Globals):
    try:
      self.items: list[AdministrableProduct] = []
      self._current_product: AdministrableProduct | None = None
      self._current_substance: Substance | None = None
      self._current_reference_substance: Substance | None = None
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME, optional=True)
      if self.success:
        for index, row in self.sheet.iterrows():
          self._create_administrable_product(index)
          self._create_ingredient_and_substance(index)
          self._create_strength(index)
          self._create_reference_substance(index)
          self._create_reference_strength(index)
    except Exception as e:
      self._sheet_exception(e)

  def _create_administrable_product(self, index):
    name = self.read_cell_by_name(index, 'name')
    if name:
      params = {
        'name': self.read_cell_by_name(index, 'name'), 
        'description': self.read_cell_by_name(index, 'description', must_be_present=False), 
        'label': self.read_cell_by_name(index, 'label', must_be_present=False),
        'administrableDoseForm': Alias(self.globals).code(self.read_cdisc_klass_attribute_cell_by_name("AdministrableProduct", "administrableDoseForm", index, "administrableDoseForm"), []),
        'pharmacologicClass': self.read_other_code_cell_by_name(index, "pharmacologicClass"),
      }
      item = self.create_object(AdministrableProduct, params)
      if item:
        self.items.append(item)
        self.globals.cross_references.add(item.name, item)
        self._current_product = item

  def _create_ingredient_and_substance(self, index):
    name = self.read_cell_by_name(index, 'substanceName')
    if name:
      params = {
        'name': name,
        'description': self.read_cell_by_name(index, 'substanceDescription', must_be_present=False),
        'label': self.read_cell_by_name(index, 'substanceLabel', must_be_present=False),
        'code': self.read_other_code_cell_by_name(index, "substanceCode")
      }
      substance = self.create_object(Substance, params)
      if substance:
        params = {
          'role': self.read_other_code_cell_by_name(index, "ingredientRole"),
          'substance': substance,
          'strengths': []
        }
        ingredient = self.create_object(Ingredient, params)
        if ingredient:
          self._current_product.ingredients.append(ingredient)
          self._current_substance = substance

  def _create_reference_substance(self, index):
    name = self.read_cell_by_name(index, 'referenceSubstanceName')
    if name:
      params = {
        'name': name,
        'description': self.read_cell_by_name(index, 'referenceSubstanceDescription', must_be_present=False),
        'label': self.read_cell_by_name(index, 'referenceSubstanceLabel', must_be_present=False),
        'code': self.read_other_code_cell_by_name(index, "referenceSubstanceCode")
      }
      substance = self.create_object(Substance, params)
      if substance:
        self._current_substance.referenceSubstance = substance
        self._current_reference_substance = substance
        return
    self._current_reference_substance = None
    return

  def _create_strength(self, index):
    params = {
      'name': self.read_cell_by_name(index, 'strengthName'), 
      'description': self.read_cell_by_name(index, 'strengthdescription', must_be_present=False), 
      'label': self.read_cell_by_name(index, 'strengthLabel', must_be_present=False),
      'numerator': self._read_numerator(index, 'strengthNumerator'),
      'denominator': self.read_quantity_cell_by_name(index, 'strengthDenominator')
    }
    strength = self.create_object(Strength, params)
    if strength:
      self._current_substance.strengths.append(strength)

  def _create_reference_strength(self, index):
    name = self.read_cell_by_name(index, 'referenceSubstanceStrengthName')
    if name:
      params = {
        'name': name, 
        'description': self.read_cell_by_name(index, 'referenceSubstanceStrengthdescription', must_be_present=False), 
        'label': self.read_cell_by_name(index, 'referenceSubstanceStrengthLabel', must_be_present=False),
        'numerator': self._read_numerator(index, 'referenceSubstanceStrengthNumerator'),
        'denominator': self.read_quantity_cell_by_name(index, 'referenceSubstanceStrengthDenominator')
      }
      strength = self.create_object(Strength, params)
      if strength:
        self._current_reference_substance.strengths.append(strength)

  def _read_numerator(self, index, field_name):
    text = self.read_cell_by_name(index, field_name)    
    return self.read_range_cell_by_name(index, field_name) if '..' in text else self.read_quantity_cell_by_name(index, field_name)


import pytest
from src.usdm_excel.globals import Globals as GlobalsClass
from tests.test_factory import Factory as FactoryClass
from tests.test_data_factory import MinimalStudy
from usdm_excel.document.elements import Elements

@pytest.fixture
def elements(mocker, factory: FactoryClass, minimal: MinimalStudy) -> Elements:
  sheet = factory.base_sheet(mocker)
  return Elements(sheet, minimal.study_version, minimal.study_definition_document_version)

def test_phase(elements):
  assert elements.study_phase() == '<usdm:ref klass="Code" id="Code_10" attribute="decode"/>'

def test_short_title(elements):
  assert elements.study_short_title() == '<usdm:ref klass="StudyTitle" id="StudyTitle_2" attribute="text"/>'

def test_full_title(elements):
  assert elements.study_full_title() == '<usdm:ref klass="StudyTitle" id="StudyTitle_1" attribute="text"/>'

def test_acronym(elements):
  assert elements.study_acronym() == '<usdm:ref klass="StudyTitle" id="StudyTitle_3" attribute="text"/>'

def test_study_rationale(elements):
  assert elements.study_rationale() == '<usdm:ref klass="StudyVersion" id="StudyVersion_1" attribute="rationale"/>'

def test_version_identifier(elements):
  assert elements.study_version_identifier() == '<usdm:ref klass="StudyVersion" id="StudyVersion_1" attribute="versionIdentifier"/>'

def test_identifier(elements):
  assert elements.study_identifier() == '<usdm:ref klass="StudyIdentifier" id="StudyIdentifier_1" attribute="text"/>'

def test_regulatory_identifiers(elements):
  assert elements.study_regulatory_identifiers() == '<usdm:ref klass="StudyIdentifier" id="StudyIdentifier_2" attribute="text"/>, <usdm:ref klass="StudyIdentifier" id="StudyIdentifier_3" attribute="text"/>'

def test_document_date(elements):
  assert elements.document_approval_date() == '<usdm:ref klass="GovernanceDate" id="GovernanceDate_2" attribute="dateValue"/>'

def test_study_date(elements):
  assert elements.study_approval_date() == '<usdm:ref klass="GovernanceDate" id="GovernanceDate_1" attribute="dateValue"/>'

def test_organization_name_and_address(elements):
  assert elements.organization_name_and_address() == '<usdm:ref klass="Organization" id="Organization_1" attribute="name"/>, <usdm:ref klass="Address" id="Address_1" attribute="text"/>'

def test_organization_address(elements):
  assert elements.organization_address() == '<usdm:ref klass="Address" id="Address_1" attribute="text"/>'

def test_organization_name(elements):
  assert elements.organization_name() == '<usdm:ref klass="Organization" id="Organization_1" attribute="name"/>'

def test_amendment(elements):
  assert elements.amendment() == '<usdm:ref klass="StudyAmendment" id="StudyAmendment_1" attribute="number"/>'

def test_amendment_scopes(elements):
  assert elements.amendment_scopes() == '<usdm:ref klass="Code" id="Code_4" attribute="decode"/>'
